import warnings 

# def custom_formatwarning(msg, *args, **kwargs):
#     # ignore everything except the message
#     return '%s:%d\n\t%s: %s\n'%(args[1],args[2],args[0].__name__,str(msg))

# warnings.formatwarning = custom_formatwarning

class DimensionError(Exception):
    """
    Class for raising error about dimension
    """


class DistributionCompatibilityError(Exception):
    """
    Class for raising error about incompatible distribution
    """


class NotYetImplemented(Exception):
    """
    Class for raising error when a component has been implemented yet
    """


class MethodImplementationError(Exception):
    """
    Class for raising error when an abstract method has not been implemented
    in the child class.
    """

    def __init__(self, subclass, method_name):
        s_f = '%s does not have an implementation of the  %s method. ' + \
            'See superclass for method description.'
        super(MethodImplementationError,self).__init__(s_f % (type(subclass).__name__, method_name))


class ParameterError(Exception):
    """
    Class for raising error about input parameters
    """


class ParameterWarning(Warning):
    """
    Class for issuing warnings about unacceptable parameters
    """


class MaxSamplesWarning(Warning):
    """
    Class for issuing warning about using maximum number of data samples
    """


class MaxLevelsWarning(Warning):
    """
    Class for issuing warning about using maximum number of data samples
    """


class CubatureWarning(Warning):
    """
    Class for issuing warnings throughout cubature algorithms
    """



from .abstract_kernel import AbstractKernelScaleLengthscales
from ..util.transforms import tf_exp_eps,tf_exp_eps_inv,tf_identity
from ..util.shift_invar_ops import BERNOULLIPOLYSDICT,bernoulli_poly
from ..util.transforms import tf_identity,tf_exp_eps,tf_exp_eps_inv
from ..util.dig_shift_invar_ops import to_bin,weighted_walsh_funcs,bin_from_numpy_to_torch,to_float
from ..util import ParameterError
import numpy as np 
from typing import Union,Tuple
import scipy.special


class AbstractSIDSIKernel(AbstractKernelScaleLengthscales):

    AUTOGRADKERNEL = False

    def __init__(self,
            d, 
            scale, 
            lengthscales,
            alpha,
            shape_alpha,
            alpha_endsize_ops,
            shape_scale,
            shape_lengthscales, 
            tfs_alpha,
            tfs_scale,
            tfs_lengthscales,
            torchify, 
            requires_grad_alpha,
            requires_grad_scale, 
            requires_grad_lengthscales, 
            device,
            compile_call,
            comiple_call_kwargs,
            ):
        input_lengthscales_is_none = lengthscales is None
        if input_lengthscales_is_none: lengthscales = 1.
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales, 
            tfs_scale = tfs_scale,
            tfs_lengthscales = tfs_lengthscales,
            torchify = torchify, 
            requires_grad_scale = requires_grad_scale, 
            requires_grad_lengthscales = requires_grad_lengthscales, 
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        self.raw_alpha = self.parse_assign_param(
            pname = "alpha",
            param = alpha, 
            shape_param = shape_alpha,
            requires_grad_param = requires_grad_alpha,
            tfs_param = tfs_alpha,
            endsize_ops = alpha_endsize_ops,
            constraints = ["POSITIVE"])
        self.tfs_alpha = tfs_alpha
        if input_lengthscales_is_none:
            lengthscales_new = (2**(1/self.d)-1.)
            raw_lengthscales_new = self.tfs_lengthscales[0](self.nptarray([lengthscales_new]))
            if self.torchify:
                self.raw_lengthscales.data[:] = raw_lengthscales_new
            else:
                self.raw_lengthscales[:] = raw_lengthscales_new
    
    @property
    def alpha(self):
        return self.tfs_alpha[1](self.raw_alpha)
    
    def parsed_single_integral_01d(self, x, batch_params):
        return batch_params["scale"][...,0]+0*x[...,0]
    
    def double_integral_01d(self):
        return self.scale[...,0]
    
    def combine_per_dim_components_raw_m1(self, kparts, beta0, beta1, c, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        ind = 1.*((beta0+beta1)==0)
        v = 1 
        p = lengthscales[...,None,:]*kparts
        v = p[...,0]
        icp = ind[...,0]
        for j in range(1,p.shape[-1]):
            v = v*(ind[...,j]+p[...,j])+icp*p[...,j]
            icp = icp*ind[...,j]
        v = v-1.*(icp==0)
        return scale*c.sum(-1),scale*(v*c).sum(-1)
    
    def combine_per_dim_components(self, kparts, beta0, beta1, c, batch_params):
        a,v = self.combine_per_dim_components_raw_m1(kparts, beta0, beta1, c, batch_params)
        k = a+v
        return k
    
    def parsed___call__(self, x0, x1, beta0, beta1, c, batch_params):
        kparts = self.get_per_dim_components(x0,x1,beta0,beta1)
        k = self.combine_per_dim_components(kparts,beta0,beta1,c,batch_params)
        return k

class KernelShiftInvar(AbstractSIDSIKernel):

    r""" 
    Shift invariant kernel with 
    smoothness $\boldsymbol{\alpha}$, product weights (lengthscales) $\boldsymbol{\gamma}$, and scale $S$:
    
    $$\begin{aligned}
        K(\boldsymbol{x},\boldsymbol{z}) &= S \prod_{j=1}^d \left(1+ \gamma_j \tilde{K}_{\alpha_j}((x_j - z_j) \mod 1))\right), \\ 
        \tilde{K}_\alpha(x) &= (-1)^{\alpha+1}\frac{(2 \pi)^{2 \alpha}}{(2\alpha)!} B_{2\alpha}(x)
    \end{aligned}$$

    where $B_n$ is the $n^\text{th}$ Bernoulli polynomial.
    
    Examples:
        >>> from qmcpy import Lattice, fftbr, ifftbr
        >>> n = 8
        >>> d = 4
        >>> lat = Lattice(d,seed=11)
        >>> x = lat(n)
        >>> x.shape
        (8, 4)
        >>> x.dtype
        dtype('float64')
        >>> kernel = KernelShiftInvar(
        ...     d = d, 
        ...     alpha = list(range(1,d+1)),
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)])
        >>> k00 = kernel(x[0],x[0])
        >>> k00.item()
        91.23444453396341
        >>> k0 = kernel(x,x[0])
        >>> with np.printoptions(precision=2):
        ...     print(k0)
        [91.23 -2.32  5.69  5.69 12.7  -4.78 -4.78 12.7 ]
        >>> assert k0[0]==k00
        >>> kmat = kernel(x[:,None,:],x[None,:,:])
        >>> with np.printoptions(precision=2):
        ...     print(kmat)
        [[91.23 -2.32  5.69  5.69 12.7  -4.78 -4.78 12.7 ]
         [-2.32 91.23  5.69  5.69 -4.78 12.7  12.7  -4.78]
         [ 5.69  5.69 91.23 -2.32 12.7  -4.78 12.7  -4.78]
         [ 5.69  5.69 -2.32 91.23 -4.78 12.7  -4.78 12.7 ]
         [12.7  -4.78 12.7  -4.78 91.23 -2.32  5.69  5.69]
         [-4.78 12.7  -4.78 12.7  -2.32 91.23  5.69  5.69]
         [-4.78 12.7  12.7  -4.78  5.69  5.69 91.23 -2.32]
         [12.7  -4.78 -4.78 12.7   5.69  5.69 -2.32 91.23]]
        >>> assert (kmat[:,0]==k0).all()
        >>> lam = np.sqrt(n)*fftbr(k0)
        >>> y = np.random.Generator(np.random.PCG64(7)).uniform(low=0,high=1,size=(n))
        >>> np.allclose(ifftbr(fftbr(y)*lam),kmat@y)
        True
        >>> np.allclose(ifftbr(fftbr(y)/lam),np.linalg.solve(kmat,y))
        True
        >>> import torch 
        >>> xtorch = torch.from_numpy(x)
        >>> kernel_torch = KernelShiftInvar(
        ...     d = d, 
        ...     alpha = list(range(1,d+1)),
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)],
        ...     torchify = True)
        >>> kmat_torch = kernel_torch(xtorch[:,None,:],xtorch[None,:,:])
        >>> np.allclose(kmat_torch.detach().numpy(),kmat)
        True
        >>> kernel.single_integral_01d(x)
        array([10., 10., 10., 10., 10., 10., 10., 10.])
        >>> kernel_torch.single_integral_01d(xtorch)
        tensor([10., 10., 10., 10., 10., 10., 10., 10.], dtype=torch.float64,
               grad_fn=<AddBackward0>)
        
        Batch Params 
        
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelShiftInvar(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,2])
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        Derivatives 

        >>> scale = rng.uniform(low=0,high=1,size=(1,))
        >>> lengthscales = rng.uniform(low=0,high=1,size=(3,))
        >>> kernel = KernelShiftInvar(
        ...     d = 3,
        ...     alpha = 3,
        ...     torchify = True,
        ...     scale = torch.from_numpy(scale),
        ...     lengthscales = torch.from_numpy(lengthscales))
        >>> x0 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> x1 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> x2 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> x = torch.stack([x0,x1,x2],axis=-1)
        >>> z0 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> z1 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> z2 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> z = torch.stack([z0,z1,z2],axis=-1)
        >>> c = torch.from_numpy(rng.uniform(low=0,high=1,size=(2,)))
        >>> beta0 = torch.tensor([
        ...     [1,0,0],
        ...     [0,2,0]])
        >>> beta1 = torch.tensor([
        ...     [0,0,2],
        ...     [2,1,0]])
        >>> with torch.no_grad():
        ...     y = kernel(x,z,beta0,beta1,c)
        >>> y
        tensor([ 3003.7945, -1517.1556,   701.6692,  1470.4241], dtype=torch.float64)
        >>> y_no_deriv = kernel(x,z)
        >>> y_first = y_no_deriv.clone()
        >>> y_first = torch.autograd.grad(y_first,x0,grad_outputs=torch.ones_like(y_first,requires_grad=True),create_graph=True)[0]
        >>> y_first = torch.autograd.grad(y_first,z2,grad_outputs=torch.ones_like(y_first,requires_grad=True),create_graph=True)[0]
        >>> y_first = torch.autograd.grad(y_first,z2,grad_outputs=torch.ones_like(y_first,requires_grad=True),create_graph=True)[0]
        >>> y_second = y_no_deriv.clone()
        >>> y_second = torch.autograd.grad(y_second,x1,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,x1,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,z0,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,z0,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,z1,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> yhat = (y_first*c[0]+y_second*c[1]).detach()
        >>> yhat
        tensor([ 3003.7946, -1517.1557,   701.6692,  1470.4241], dtype=torch.float64)
        >>> torch.allclose(y,yhat)
        True
        >>> kernel = KernelShiftInvar(
        ...     d = 3,
        ...     alpha = 3,
        ...     scale = scale,
        ...     lengthscales = lengthscales)
        >>> ynp = kernel(x.detach().numpy(),z.detach().numpy(),beta0.numpy(),beta1.numpy(),c.numpy())
        >>> ynp
        array([ 3003.79938927, -1517.15808005,   701.6700331 ,  1470.42580601])
        >>> np.allclose(ynp,y.numpy())
        True
        
    **References:** 
    
    1.  Kaarnioja, Vesa, Frances Y. Kuo, and Ian H. Sloan.  
        "Lattice-based kernel approximation and serendipitous weights for parametric PDEs in very high dimensions."  
        International Conference on Monte Carlo and Quasi-Monte Carlo Methods in Scientific Computing. Cham: Springer International Publishing, 2022.
    """
        
    def __init__(self,
            d, 
            scale = 1., 
            lengthscales = None,
            alpha = 2,
            shape_scale = [1],
            shape_lengthscales = None, 
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False, 
            requires_grad_scale = True, 
            requires_grad_lengthscales = True, 
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Product weights $(\gamma_1,\dots,\gamma_d)$.
            alpha (Union[np.ndarray,torch.Tensor]): Smoothness parameters $(\alpha_1,\dots,\alpha_d)$ where $\alpha_j \geq 1$ for $j=1,\dots,d$.
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            alpha = alpha, 
            shape_alpha = [d], 
            alpha_endsize_ops = [d],
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales, 
            tfs_alpha = (tf_identity,tf_identity), 
            tfs_scale = tfs_scale,
            requires_grad_alpha = False,
            tfs_lengthscales = tfs_lengthscales, 
            torchify = torchify,
            requires_grad_scale = requires_grad_scale,
            requires_grad_lengthscales = requires_grad_lengthscales,
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        assert self.alpha.shape==(self.d,)
        assert all(int(alphaj) in BERNOULLIPOLYSDICT for alphaj in self.alpha)
        if self.torchify:
            import torch 
            self.lgamma = torch.lgamma 
        else:
            self.lgamma = scipy.special.loggamma
    
    def get_per_dim_components(self, x0, x1, beta0, beta1):
        p = len(beta0)
        betasum = beta0+beta1
        order = 2*self.alpha-betasum
        assert (2<=order).all(), "order must all be at least 2, but got order = %s"%str(order)
        coeffs = (-1)**(self.alpha+beta1+1)*self.npt.exp(2*self.alpha*np.log(2*np.pi)-self.lgamma(order+1))
        delta = (x0-x1)%1
        kperdim = coeffs*self.npt.stack([self.npt.concatenate([bernoulli_poly(int(order[l,j].item()),delta[...,j,None]) for j in range(self.d)],-1) for l in range(p)],-2)
        return kperdim


class KernelShiftInvarCombined(AbstractSIDSIKernel):

    r""" 
    Shift invariant kernel with 
    combination weights $\boldsymbol{\alpha}_1,\dots,\boldsymbol{\alpha}_d \in \mathbb{R}_{>0}^4$, product weights (lengthscales) $\boldsymbol{\gamma}$, and scale $S$:
    
    $$\begin{aligned}
        K(\boldsymbol{x},\boldsymbol{z}) &= S \prod_{j=1}^d \left(1+ \gamma_j \left(\sum_{p=1}^4 \alpha_{jp} \tilde{K}_p(x_j \mod 1 z_j)\right)\right)
    \end{aligned}$$

    where, $\tilde{K}_p$ are defined in `KernelShiftInvar` for $p \in \{1,2,3,4\}$
    
    Examples:
        >>> from qmcpy import Lattice, fftbr, ifftbr
        >>> n = 8
        >>> d = 4
        >>> lat = Lattice(d,seed=11)
        >>> x = lat(n)
        >>> x.shape
        (8, 4)
        >>> x.dtype
        dtype('float64')
        >>> kernel = KernelShiftInvarCombined(
        ...     d = d, 
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)])
        >>> k00 = kernel(x[0],x[0])
        >>> k00.item()
        117.22427096475315
        >>> k0 = kernel(x,x[0])
        >>> with np.printoptions(precision=2):
        ...     print(k0)
        [ 117.22  153.84   28.59   36.1  -139.83   77.1    -8.51   35.22]
        >>> assert k0[0]==k00
        >>> kmat = kernel(x[:,None,:],x[None,:,:])
        >>> with np.printoptions(precision=2):
        ...     print(kmat)
        [[ 117.22  153.84   36.1    28.59   35.22   -8.51   77.1  -139.83]
         [ 153.84  117.22   28.59   36.1    -8.51   35.22 -139.83   77.1 ]
         [  28.59   36.1   117.22  153.84 -139.83   77.1    35.22   -8.51]
         [  36.1    28.59  153.84  117.22   77.1  -139.83   -8.51   35.22]
         [-139.83   77.1    35.22   -8.51  117.22  153.84   36.1    28.59]
         [  77.1  -139.83   -8.51   35.22  153.84  117.22   28.59   36.1 ]
         [  -8.51   35.22 -139.83   77.1    28.59   36.1   117.22  153.84]
         [  35.22   -8.51   77.1  -139.83   36.1    28.59  153.84  117.22]]
        >>> assert (kmat[:,0]==k0).all()
        >>> lam = np.sqrt(n)*fftbr(k0)
        >>> y = np.random.Generator(np.random.PCG64(7)).uniform(low=0,high=1,size=(n))
        >>> np.allclose(ifftbr(fftbr(y)*lam),kmat@y)
        True
        >>> np.allclose(ifftbr(fftbr(y)/lam),np.linalg.solve(kmat,y))
        True
        >>> import torch 
        >>> xtorch = torch.from_numpy(x)
        >>> kernel_torch = KernelShiftInvarCombined(
        ...     d = d, 
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)],
        ...     torchify = True)
        >>> kmat_torch = kernel_torch(xtorch[:,None,:],xtorch[None,:,:])
        >>> np.allclose(kmat_torch.detach().numpy(),kmat)
        True
        >>> kernel.single_integral_01d(x)
        array([10., 10., 10., 10., 10., 10., 10., 10.])
        >>> kernel_torch.single_integral_01d(xtorch)
        tensor([10., 10., 10., 10., 10., 10., 10., 10.], dtype=torch.float64,
               grad_fn=<AddBackward0>)
        
        Batch Params 
        
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelShiftInvarCombined(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,2])
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)
        
    **References:** 
    
    1.  Kaarnioja, Vesa, Frances Y. Kuo, and Ian H. Sloan.  
        "Lattice-based kernel approximation and serendipitous weights for parametric PDEs in very high dimensions."  
        International Conference on Monte Carlo and Quasi-Monte Carlo Methods in Scientific Computing. Cham: Springer International Publishing, 2022.
    """
        
    def __init__(self,
            d, 
            scale = 1., 
            lengthscales = None,
            alpha = 1,
            shape_scale = [1],
            shape_lengthscales = None, 
            shape_alpha = None,
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            tfs_alpha = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False, 
            requires_grad_scale = True, 
            requires_grad_lengthscales = True, 
            requires_grad_alpha = True,
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Product weights $(\gamma_1,\dots,\gamma_d)$.
            alpha (Union[np.ndarray,torch.Tensor]): Weights $\boldsymbol{\alpha}_1,\dots,\boldsymbol{\alpha}_d \in \mathbb{R}_{>0}^4$. 
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            shape_alpha (list): Shape of `alpha` when `np.isscalar(alpha)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_alpha (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            requires_grad_alpha (bool): If `True` and `torchify`, set `requires_grad=True` for `alpha`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            alpha = alpha, 
            alpha_endsize_ops = [d],
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales, 
            shape_alpha = [4,d] if shape_alpha is None else shape_alpha, 
            tfs_scale = tfs_scale,
            tfs_alpha = tfs_alpha, 
            tfs_lengthscales = tfs_lengthscales, 
            torchify = torchify,
            requires_grad_scale = requires_grad_scale,
            requires_grad_lengthscales = requires_grad_lengthscales,
            requires_grad_alpha = requires_grad_alpha,
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        assert self.alpha.shape[-2:]==(4,d)
        if self.torchify:
            import torch 
            self.lgamma = torch.lgamma 
        else:
            self.lgamma = scipy.special.loggamma
        a = self.npt.arange(1,5,**self.nptkwargs)
        self.coeffs = (-1)**(a+1)*self.npt.exp(2*a*np.log(2*np.pi)-self.lgamma(2*a+1))
    
    def get_per_dim_components(self, x0, x1, beta0, beta1):
        p = len(beta0)
        assert (beta0==0).all() and (beta1==0).all(), "KernelDSICombined does not support derivatives"
        delta = (x0-x1)%1
        kparts = [None]*4 
        kparts[0] = bernoulli_poly(1,delta)
        kparts[1] = bernoulli_poly(2,delta)
        kparts[2] = bernoulli_poly(3,delta)
        kparts[3] = bernoulli_poly(4,delta)
        kparts = self.coeffs[:,None]*self.npt.stack(kparts,-2)
        kperdim = self.npt.stack([kparts for j in range(p)],-2)
        return kperdim

    def combine_per_dim_components_raw_m1(self, kparts, beta0, beta1, c, batch_params):
        kparts = (self.alpha[...,None,:,None,:]*kparts).sum(-3)
        return super().combine_per_dim_components_raw_m1(kparts,beta0,beta1,c,batch_params)

class KernelDigShiftInvar(AbstractSIDSIKernel):
    
    r""" 
    Digitally shift invariant kernel in base $b=2$ with 
    smoothness $\boldsymbol{\alpha}$, product weights $\boldsymbol{\gamma}$, and scale $S$: 
    
    $$\begin{aligned}
        K(\boldsymbol{x},\boldsymbol{z}) &= S \prod_{j=1}^d \left(1+ \gamma_j \tilde{K}_{\alpha_j}(x_j \oplus z_j)\right), \qquad\mathrm{where} \\
        \tilde{K}_1(x) &= 6 \left(\frac{1}{6} - 2^{\lfloor \log_2(x) \rfloor -1}\right), \\
        \tilde{K}_2(x) &= \sum_{k \in \mathbb{N}} \frac{\mathrm{wal}_k(x)}{2^{\mu_2(k)}} = -\beta(x) x + \frac{5}{2}\left[1-t_1(x)\right]-1, \\
        \tilde{K}_3(x) &= \sum_{k \in \mathbb{N}} \frac{\mathrm{wal}_k(x)}{2^{\mu_3(k)}} = \beta(x)x^2-5\left[1-t_1(x)\right]x+\frac{43}{18}\left[1-t_2(x)\right]-1, \\
        \tilde{K}_4(x) &= \sum_{k \in \mathbb{N}} \frac{\mathrm{wal}_k(x)}{2^{\mu_4(k)}} = - \frac{2}{3}\beta(x)x^3+5\left[1-t_1(x)\right]x^2 - \frac{43}{9}\left[1-t_2(x)\right]x +\frac{701}{294}\left[1-t_3(x)\right]+\beta(x)\left[\frac{1}{48}\sum_{a=0}^\infty \frac{\mathrm{wal}_{2^a}(x)}{2^{3a}} - \frac{1}{42}\right] - 1.
    \end{aligned}$$

    where 
    
    - $x \oplus z$ is XOR between bits, 
    - $\mathrm{wal}_k$ is the $k^\text{th}$ Walsh function, 
    - $\beta(x) = - \lfloor \log_2(x) \rfloor$ and $t_\nu(x) = 2^{-\nu \beta(x)}$ where $\beta(0)=t_\nu(0) = 0$, and 
    - and $\mu_\alpha$ is the Dick weight function which sums the first $\alpha$ largest indices of $1$ bits in the binary expansion of $k$ 
    e.g. $k=13=1101_2$ has 1-bit indexes $(4,3,1)$ so 
    
    $$\mu_1(k) = 4, \mu_2(k) = 4+3, \mu_3(k) = 4+3+1 = \mu_4(k) = \mu_5(k) = \dots.$$

    Examples:
        >>> from qmcpy import DigitalNetB2, fwht
        >>> n = 8
        >>> d = 4
        >>> dnb2 = DigitalNetB2(d,seed=11)
        >>> x = dnb2(n,return_binary=True)
        >>> x.shape
        (8, 4)
        >>> x.dtype
        dtype('uint64')
        >>> kernel = KernelDigShiftInvar(
        ...     d = d, 
        ...     t = dnb2.t,
        ...     alpha = list(range(1,d+1)),
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)])
        >>> k00 = kernel(x[0],x[0])
        >>> k00.item()
        34.49037002918452
        >>> k0 = kernel(x,x[0])
        >>> with np.printoptions(precision=2):
        ...     print(k0)
        [34.49  4.15  9.59  4.98 15.42  5.45 11.99  4.51]
        >>> assert k0[0]==k00
        >>> kmat = kernel(x[:,None,:],x[None,:,:])
        >>> with np.printoptions(precision=2):
        ...     print(kmat)
        [[34.49  4.15  9.59  4.98 15.42  5.45 11.99  4.51]
         [ 4.15 34.49  4.98  9.59  5.45 15.42  4.51 11.99]
         [ 9.59  4.98 34.49  4.15 11.99  4.51 15.42  5.45]
         [ 4.98  9.59  4.15 34.49  4.51 11.99  5.45 15.42]
         [15.42  5.45 11.99  4.51 34.49  4.15  9.59  4.98]
         [ 5.45 15.42  4.51 11.99  4.15 34.49  4.98  9.59]
         [11.99  4.51 15.42  5.45  9.59  4.98 34.49  4.15]
         [ 4.51 11.99  5.45 15.42  4.98  9.59  4.15 34.49]]
        >>> assert (kmat[:,0]==k0).all()
        >>> lam = np.sqrt(n)*fwht(k0)
        >>> y = np.random.Generator(np.random.PCG64(7)).uniform(low=0,high=1,size=(n))
        >>> np.allclose(fwht(fwht(y)*lam),kmat@y)
        True
        >>> np.allclose(fwht(fwht(y)/lam),np.linalg.solve(kmat,y))
        True
        >>> import torch 
        >>> xtorch = bin_from_numpy_to_torch(x)
        >>> kernel_torch = KernelDigShiftInvar(
        ...     d = d, 
        ...     t = dnb2.t,
        ...     alpha = list(range(1,d+1)),
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)],
        ...     torchify = True)
        >>> kmat_torch = kernel_torch(xtorch[:,None,:],xtorch[None,:,:])
        >>> np.allclose(kmat_torch.detach().numpy(),kmat)
        True
        >>> xf = to_float(x,dnb2.t)
        >>> kmat_from_floats = kernel(xf[:,None,:],xf[None,:,:])
        >>> np.allclose(kmat,kmat_from_floats)
        True
        >>> xftorch = to_float(xtorch,dnb2.t)
        >>> xftorch.dtype
        torch.float32
        >>> kmat_torch_from_floats = kernel_torch(xftorch[:,None,:],xftorch[None,:,:])
        >>> torch.allclose(kmat_torch_from_floats,kmat_torch)
        True
        >>> kernel.single_integral_01d(x)
        array([10., 10., 10., 10., 10., 10., 10., 10.])
        >>> kernel_torch.single_integral_01d(xtorch)
        tensor([10., 10., 10., 10., 10., 10., 10., 10.], grad_fn=<AddBackward0>)

        Batch Params 
        
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelDigShiftInvar(
        ...     d = 2, 
        ...     t = 10,
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,2])
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

    **References:**
        
    1.  Dick, Josef.  
        "Walsh spaces containing smooth functions and quasi-Monte Carlo rules of arbitrary high order."  
        SIAM Journal on Numerical Analysis 46.3 (2008): 1519-1553.

    2.  Dick, Josef.  
        "The decay of the Walsh coefficients of smooth functions."  
        Bulletin of the Australian Mathematical Society 80.3 (2009): 430-453.  

    3.  Jagadeeswaran, Rathinavel, and Fred J. Hickernell.  
        "Fast automatic Bayesian cubature using Sobol' sampling."  
        Advances in Modeling and Simulation: Festschrift for Pierre L'Ecuyer. Cham: Springer International Publishing, 2022. 301-318.

    4.  Rathinavel, Jagadeeswaran.  
        Fast automatic Bayesian cubature using matching kernels and designs.  
        Illinois Institute of Technology, 2019.
    
    5.  Sorokin, Aleksei.  
        "A Unified Implementation of Quasi-Monte Carlo Generators, Randomization Routines, and Fast Kernel Methods."  
        arXiv preprint arXiv:2502.14256 (2025).
    """
        
    def __init__(self,
            d,
            t = None,
            scale = 1., 
            lengthscales = None,
            alpha = 2,
            shape_scale = [1],
            shape_lengthscales = None, 
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False, 
            requires_grad_scale = True, 
            requires_grad_lengthscales = True, 
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            t (int): number of bits in binary represtnations. Typically `dnb2.t` where `isinstance(dnb2,DigitalNetB2)`.
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Product weights $(\gamma_1,\dots,\gamma_d)$.
            alpha (Union[np.ndarray,torch.Tensor]): Smoothness parameters $(\alpha_1,\dots,\alpha_d)$ where $\alpha_j \geq 1$ for $j=1,\dots,d$.
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            alpha = alpha,
            shape_alpha = [d], 
            alpha_endsize_ops = [d],
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales, 
            tfs_alpha = (tf_identity,tf_identity), 
            tfs_scale = tfs_scale, 
            tfs_lengthscales = tfs_lengthscales, 
            torchify = torchify,
            requires_grad_alpha = False,
            requires_grad_scale = requires_grad_scale,
            requires_grad_lengthscales = requires_grad_lengthscales,
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        assert self.alpha.shape==(self.d,)
        self.set_t(t)
        assert all(1<=int(alphaj)<=4 for alphaj in self.alpha)
        
    
    @property
    def t(self):
        if self._t is None: 
            raise ParameterError("please use set_t to set the t value")
        return self._t
    
    def set_t(self, t):
        if t is None: 
            self._t = t
        else:
            assert t%1==0
            if self.torchify:
                assert 0<=t<=63 # torch only supports torch.int64
            else:
                assert 0<=t<=64 # numpy supports np.uint64
            self._t = t

    def get_per_dim_components(self, x0, x1, beta0, beta1):
        t = self.t
        x0 = to_bin(x0,t)
        x1 = to_bin(x1,t)
        p = len(beta0)
        betasum = beta0+beta1
        order = self.alpha-betasum
        assert (1<=order).all() and (order<=4).all(), "order must all be between 2 and 4, but got order = %s. Try increasing alpha"%str(order)
        assert not ((order==1)*(self.alpha>1)).any(), "taking the derivative of the order 2 digitally shift invariant kernel is not supported"
        ind = 1.*(betasum>0)
        delta = x0^x1 
        kparts = [None]*p
        for l in range(p):
            kparts_l = [None]*self.d
            for j in range(self.d):
                deltaj = delta[...,j,None]
                if order[l,j]==1: # order[j]=alpha[j] as we cannot take derivatives WRT the alpha=1 kernel and this cannot be the derivative of any kernels
                    flog2deltaj = -self.npt.inf*self.npt.ones(deltaj.shape,**self.nptkwargs)
                    pos = deltaj>0 
                    flog2deltaj[pos] = self.npt.floor(self.npt.log2(deltaj[pos]))-t
                    kparts_l[j] = 6*(1/6-2**(flog2deltaj-1))
                else:
                    kparts_l[j] = weighted_walsh_funcs(int(order[l,j]),deltaj[...,None],t)[...,0]-1
            kparts[l] = self.npt.concatenate(kparts_l,-1)
        kparts = self.npt.stack(kparts,-2)
        kperdim = (-2)**betasum*(ind+kparts)
        return kperdim

class KernelDigShiftInvarAdaptiveAlpha(AbstractSIDSIKernel):
    
    r""" 
    Digitally shift invariant kernel in base $b=2$ with 
    smoothness $\boldsymbol{\alpha} \geq \boldsymbol{0}$, product weights $\boldsymbol{\gamma}$, and scale $S$: 
    
    $$\begin{aligned}
        K(\boldsymbol{x},\boldsymbol{z}) &= S \prod_{j=1}^d \left(1+ \gamma_j \tilde{K}_{\alpha_j}(x_j \oplus z_j)\right), \qquad\mathrm{where} \\
        \tilde{K}_\alpha(x) &= \sum_{k \in \mathbb{N}} \frac{\mathrm{wal}_k(x)}{2^{{\alpha+1} (\mu_1(k)-1)}} = \frac{2^{\alpha+1}}{2^{\alpha+1}-2} - \left(\frac{2^{\alpha+1}}{2^{\alpha+1}-2}+1\right) 2^{\alpha(\lfloor \log_2(x) \rfloor+1)}, \\
    \end{aligned}$$

    where 
    
    - $x \oplus z$ is XOR between bits, 
    - $\mathrm{wal}_k$ is the $k^\text{th}$ Walsh function, 
    - $\beta(x) = - \lfloor \log_2(x) \rfloor$ and $t_\nu(x) = 2^{-\nu \beta(x)}$ where $\beta(0)=t_\nu(0) = 0$, and 
    - and $\mu_\alpha$ is the Dick weight function which sums the first $\alpha$ largest indices of $1$ bits in the binary expansion of $k$ 
    e.g. $k=13=1101_2$ has 1-bit indexes $(4,3,1)$ so 
    
    $$\mu_1(k) = 4, \mu_2(k) = 4+3, \mu_3(k) = 4+3+1 = \mu_4(k) = \mu_5(k) = \dots.$$

    Examples:
        >>> from qmcpy import DigitalNetB2, fwht
        >>> n = 8
        >>> d = 4
        >>> dnb2 = DigitalNetB2(d,seed=11)
        >>> x = dnb2(n,return_binary=True)
        >>> x.shape
        (8, 4)
        >>> x.dtype
        dtype('uint64')
        >>> kernel = KernelDigShiftInvarAdaptiveAlpha(
        ...     d = d, 
        ...     t = dnb2.t,
        ...     alpha = list(range(1,d+1)),
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)])
        >>> k00 = kernel(x[0],x[0])
        >>> k00.item()
        48.08465608465609
        >>> k0 = kernel(x,x[0])
        >>> with np.printoptions(precision=2):
        ...     print(k0)
        [48.08  0.    9.38  0.   19.74  0.   14.84  0.  ]
        >>> assert k0[0]==k00
        >>> kmat = kernel(x[:,None,:],x[None,:,:])
        >>> with np.printoptions(precision=2):
        ...     print(kmat)
        [[48.08  0.    9.38  0.   19.74  0.   14.84  0.  ]
         [ 0.   48.08  0.    9.38  0.   19.74  0.   14.84]
         [ 9.38  0.   48.08  0.   14.84  0.   19.74  0.  ]
         [ 0.    9.38  0.   48.08  0.   14.84  0.   19.74]
         [19.74  0.   14.84  0.   48.08  0.    9.38  0.  ]
         [ 0.   19.74  0.   14.84  0.   48.08  0.    9.38]
         [14.84  0.   19.74  0.    9.38  0.   48.08  0.  ]
         [ 0.   14.84  0.   19.74  0.    9.38  0.   48.08]]
        >>> assert (kmat[:,0]==k0).all()
        >>> lam = np.sqrt(n)*fwht(k0)
        >>> y = np.random.Generator(np.random.PCG64(7)).uniform(low=0,high=1,size=(n))
        >>> np.allclose(fwht(fwht(y)*lam),kmat@y)
        True
        >>> np.allclose(fwht(fwht(y)/lam),np.linalg.solve(kmat,y))
        True
        >>> import torch 
        >>> xtorch = bin_from_numpy_to_torch(x)
        >>> kernel_torch = KernelDigShiftInvarAdaptiveAlpha(
        ...     d = d, 
        ...     t = dnb2.t,
        ...     alpha = list(range(1,d+1)),
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)],
        ...     torchify = True)
        >>> kmat_torch = kernel_torch(xtorch[:,None,:],xtorch[None,:,:])
        >>> np.allclose(kmat_torch.detach().numpy(),kmat,atol=1e-5)
        True
        >>> xf = to_float(x,dnb2.t)
        >>> kmat_from_floats = kernel(xf[:,None,:],xf[None,:,:])
        >>> np.allclose(kmat,kmat_from_floats)
        True
        >>> xftorch = to_float(xtorch,dnb2.t)
        >>> xftorch.dtype
        torch.float32
        >>> kmat_torch_from_floats = kernel_torch(xftorch[:,None,:],xftorch[None,:,:])
        >>> torch.allclose(kmat_torch_from_floats,kmat_torch)
        True
        >>> kernel.single_integral_01d(x)
        array([10., 10., 10., 10., 10., 10., 10., 10.])
        >>> kernel_torch.single_integral_01d(xtorch)
        tensor([10., 10., 10., 10., 10., 10., 10., 10.], grad_fn=<AddBackward0>)

        Batch Params 
        
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelDigShiftInvarAdaptiveAlpha(
        ...     d = 2, 
        ...     t = 10,
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,2])
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

    **References:**
        
    3.  Dick, Josef, and Friedrich Pillichshammer.  
        "Multivariate integration in weighted Hilbert spaces based on Walsh functions and weighted Sobolev spaces."  
        Journal of Complexity 21.2 (2005): 149-195.
    """
        
    def __init__(self,
            d,
            t=None,
            scale = 1., 
            lengthscales = None,
            alpha = 1,
            shape_scale = [1],
            shape_lengthscales = None, 
            shape_alpha = None,
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            tfs_alpha = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False, 
            requires_grad_scale = True, 
            requires_grad_lengthscales = True, 
            requires_grad_alpha = True, 
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            t (int): number of bits in binary represtnations. Typically `dnb2.t` where `isinstance(dnb2,DigitalNetB2)`.
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Product weights $(\gamma_1,\dots,\gamma_d)$.
            alpha (Union[np.ndarray,torch.Tensor]): Smoothness parameters $(\alpha_1,\dots,\alpha_d)$ where $\alpha_j \geq 1$ for $j=1,\dots,d$.
            shape_alpha (list): Shape of `alpha` when `np.isscalar(alpha)`
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_alpha (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            requires_grad_alpha (bool): If `True` and `torchify`, set `requires_grad=True` for `alpha`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            alpha = alpha, 
            shape_alpha = [d] if shape_alpha is None else shape_alpha,
            alpha_endsize_ops = [d],
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales,
            tfs_alpha = tfs_alpha, 
            tfs_scale = tfs_scale, 
            tfs_lengthscales = tfs_lengthscales, 
            torchify = torchify,
            requires_grad_alpha = requires_grad_alpha,
            requires_grad_scale = requires_grad_scale,
            requires_grad_lengthscales = requires_grad_lengthscales,
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        self.set_t(t)
        self.batch_param_names.append("alpha")
    
    @property
    def t(self):
        if self._t is None: 
            raise ParameterError("please use set_t to set the t value")
        return self._t
    
    def set_t(self, t):
        if t is None: 
            self._t = t
        else:
            assert t%1==0
            if self.torchify:
                assert 0<=t<=63 # torch only supports torch.int64
            else:
                assert 0<=t<=64 # numpy supports np.uint64
            self._t = t

    def get_per_dim_components(self, x0, x1, beta0, beta1):
        t = self.t
        x0 = to_bin(x0,t)
        x1 = to_bin(x1,t)
        assert (beta0==0).all() and (beta1==0).all(), "KernelDigShiftInvarAdaptiveAlpha does not support taking derivatives"
        p = len(beta0)
        delta = x0^x1
        flog2delta = self.npt.zeros(delta.shape,**self.nptkwargs) # should be -inf
        pos = delta>0 
        flog2delta[pos] = self.npt.floor(self.npt.log2(delta[pos]))-t
        flog2deltas = self.npt.stack([flog2delta for j in range(p)],-2)
        return flog2deltas
    
    def combine_per_dim_components_raw_m1(self, flog2deltas, beta0, beta1, c, batch_params):
        alpha = batch_params["alpha"]
        p2alphap1 = 2**(alpha+1)
        nu = p2alphap1/(p2alphap1-2)
        neginfs = flog2deltas==0
        s = (nu+1)*2**(alpha*(flog2deltas+1))
        kparts = nu-s
        kparts[neginfs] = (nu-0*s)[neginfs]
        return super().combine_per_dim_components_raw_m1(kparts,beta0,beta1,c,batch_params)
    
class KernelDigShiftInvarCombined(AbstractSIDSIKernel):
    
    r""" 
    Digitally shift invariant kernel in base $b=2$ with 
    combination weights $\boldsymbol{\alpha}_1,\dots,\boldsymbol{\alpha}_d \in \mathbb{R}_{>0}^4$, smoothness $\boldsymbol{\alpha}$, product weights $\boldsymbol{\gamma}$, and scale $S$: 
    
    $$\begin{aligned}
        K(\boldsymbol{x},\boldsymbol{z}) &= S \prod_{j=1}^d \left(1+ \gamma_j \left(\sum_{p=1}^4 \alpha_{jp} \tilde{K}_p(x_j \oplus z_j)\right)\right)
    \end{aligned}$$

    where, $\oplus$ is defined in the docs for `KernelDigShiftInvar` and so are $\tilde{K}_p$ for $p \in \{1,2,3,4\}$

    Examples:
        >>> from qmcpy import DigitalNetB2, fwht
        >>> n = 8
        >>> d = 4
        >>> dnb2 = DigitalNetB2(d,seed=11)
        >>> x = dnb2(n,return_binary=True)
        >>> x.shape
        (8, 4)
        >>> x.dtype
        dtype('uint64')
        >>> kernel = KernelDigShiftInvarCombined(
        ...     d = d, 
        ...     t = dnb2.t,
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)])
        >>> k00 = kernel(x[0],x[0])
        >>> k00.item()
        306.66030731930863
        >>> k0 = kernel(x,x[0])
        >>> with np.printoptions(precision=2):
        ...     print(k0)
        [306.66  -1.65   6.88 -13.5   22.57  -9.34   7.99  -7.54]
        >>> assert k0[0]==k00
        >>> kmat = kernel(x[:,None,:],x[None,:,:])
        >>> with np.printoptions(precision=2):
        ...     print(kmat)
        [[306.66  -1.65   6.88 -13.5   22.57  -9.34   7.99  -7.54]
         [ -1.65 306.66 -13.5    6.88  -9.34  22.57  -7.54   7.99]
         [  6.88 -13.5  306.66  -1.65   7.99  -7.54  22.57  -9.34]
         [-13.5    6.88  -1.65 306.66  -7.54   7.99  -9.34  22.57]
         [ 22.57  -9.34   7.99  -7.54 306.66  -1.65   6.88 -13.5 ]
         [ -9.34  22.57  -7.54   7.99  -1.65 306.66 -13.5    6.88]
         [  7.99  -7.54  22.57  -9.34   6.88 -13.5  306.66  -1.65]
         [ -7.54   7.99  -9.34  22.57 -13.5    6.88  -1.65 306.66]]
        >>> assert (kmat[:,0]==k0).all()
        >>> lam = np.sqrt(n)*fwht(k0)
        >>> y = np.random.Generator(np.random.PCG64(7)).uniform(low=0,high=1,size=(n))
        >>> np.allclose(fwht(fwht(y)*lam),kmat@y)
        True
        >>> np.allclose(fwht(fwht(y)/lam),np.linalg.solve(kmat,y))
        True
        >>> import torch 
        >>> xtorch = bin_from_numpy_to_torch(x)
        >>> kernel_torch = KernelDigShiftInvarCombined(
        ...     d = d, 
        ...     t = dnb2.t,
        ...     scale = 10,
        ...     lengthscales = [1/j**2 for j in range(1,d+1)],
        ...     torchify = True)
        >>> kmat_torch = kernel_torch(xtorch[:,None,:],xtorch[None,:,:])
        >>> np.allclose(kmat_torch.detach().numpy(),kmat)
        True
        >>> xf = to_float(x,dnb2.t)
        >>> kmat_from_floats = kernel(xf[:,None,:],xf[None,:,:])
        >>> np.allclose(kmat,kmat_from_floats)
        True
        >>> xftorch = to_float(xtorch,dnb2.t)
        >>> xftorch.dtype
        torch.float32
        >>> kmat_torch_from_floats = kernel_torch(xftorch[:,None,:],xftorch[None,:,:])
        >>> torch.allclose(kmat_torch_from_floats,kmat_torch)
        True
        >>> kernel.single_integral_01d(x)
        array([10., 10., 10., 10., 10., 10., 10., 10.])
        >>> kernel_torch.single_integral_01d(xtorch)
        tensor([10., 10., 10., 10., 10., 10., 10., 10.], grad_fn=<AddBackward0>)

        Batch Params 
        
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelDigShiftInvarCombined(
        ...     d = 2, 
        ...     t = 10,
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,2])
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)
    """
        
    def __init__(self,
            d,
            t = None,
            scale = 1., 
            lengthscales = None,
            alpha = 1.,
            shape_scale = [1],
            shape_lengthscales = None,
            shape_alpha = None,
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            tfs_alpha = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False, 
            requires_grad_scale = True, 
            requires_grad_lengthscales = True,
            requires_grad_alpha = True,
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            t (int): number of bits in binary represtnations. Typically `dnb2.t` where `isinstance(dnb2,DigitalNetB2)`.
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Product weights $(\gamma_1,\dots,\gamma_d)$.
            alpha (Union[np.ndarray,torch.Tensor]): Weights $\boldsymbol{\alpha}_1,\dots,\boldsymbol{\alpha}_d \in \mathbb{R}_{>0}^4$. 
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            shape_alpha (list): Shape of `alpha` when `np.isscalar(alpha)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            requires_grad_alpha (bool): If `True` and `torchify`, set `requires_grad=True` for `alpha`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            alpha = alpha,
            shape_alpha = [4,d] if shape_alpha is None else shape_alpha, 
            alpha_endsize_ops = [d],
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales, 
            tfs_alpha = tfs_alpha, 
            tfs_scale = tfs_scale, 
            tfs_lengthscales = tfs_lengthscales, 
            torchify = torchify,
            requires_grad_alpha = requires_grad_alpha,
            requires_grad_scale = requires_grad_scale,
            requires_grad_lengthscales = requires_grad_lengthscales,
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        self.set_t(t)
        assert self.alpha.shape[-2:]==(4,d)
    
    @property
    def t(self):
        if self._t is None: 
            raise ParameterError("please use set_t to set the t value")
        return self._t
    
    def set_t(self, t):
        if t is None: 
            self._t = t
        else:
            assert t%1==0
            if self.torchify:
                assert 0<=t<=63 # torch only supports torch.int64
            else:
                assert 0<=t<=64 # numpy supports np.uint64
            self._t = t

    def get_per_dim_components(self, x0, x1, beta0, beta1):
        t = self.t
        x0 = to_bin(x0,t)
        x1 = to_bin(x1,t)
        p = len(beta0)
        assert (beta0==0).all() and (beta1==0).all(), "KernelDSICombined does not support derivatives"
        delta = x0^x1
        kparts = [None]*4 
        flog2deltaj = -self.npt.inf*self.npt.ones(delta.shape,**self.nptkwargs)
        pos = delta>0 
        flog2deltaj[pos] = self.npt.floor(self.npt.log2(delta[pos]))-t
        kparts[0] = 6*(1/6-2**(flog2deltaj-1))
        kparts[1] = weighted_walsh_funcs(2,delta,t)-1
        kparts[2] = weighted_walsh_funcs(3,delta,t)-1
        kparts[3] = weighted_walsh_funcs(4,delta,t)-1
        kparts = self.npt.stack(kparts,-2)
        kperdim = self.npt.stack([kparts for j in range(p)],-2)
        return kperdim
    
    def combine_per_dim_components_raw_m1(self, kparts, beta0, beta1, c, batch_params):
        kparts = (self.alpha[...,None,:,None,:]*kparts).sum(-3)
        return super().combine_per_dim_components_raw_m1(kparts,beta0,beta1,c,batch_params)



from .abstract_stopping_criterion import AbstractStoppingCriterion
from ..util.data import Data
from ..discrete_distribution.abstract_discrete_distribution import AbstractDiscreteDistribution
from ..discrete_distribution import Lattice,DigitalNetB2,Halton
from ..discrete_distribution.abstract_discrete_distribution import AbstractLDDiscreteDistribution
from ..true_measure import Gaussian,Uniform
from ..integrand.keister import Keister
from ..integrand.box_integral import BoxIntegral
from ..integrand.sensitivity_indices import SensitivityIndices
from ..integrand.genz import Genz
from ..util import MaxSamplesWarning, NotYetImplemented, ParameterWarning, ParameterError
import numpy as np
from scipy.stats import t
from time import time
import warnings


class CubQMCRepStudentT(AbstractStoppingCriterion):
    r"""
    Quasi-Monte Carlo stopping criterion based on Student's $t$-distribution for multiple replications.
    
    Examples:
        >>> k = Keister(DigitalNetB2(seed=7,replications=25))
        >>> sc = CubQMCRepStudentT(k,abs_tol=1e-3,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> solution
        array(1.3803927)
        >>> data
        Data (Data)
            solution        1.380
            comb_bound_low  1.380
            comb_bound_high 1.381
            comb_bound_diff 0.002
            comb_flags      1
            n_total         6400
            n               6400
            n_rep           2^(8)
            time_integrate  ...
        CubQMCRepStudentT (AbstractStoppingCriterion)
            inflate         1
            alpha           0.010
            abs_tol         0.001
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(30)
        Keister (AbstractIntegrand)
        Gaussian (AbstractTrueMeasure)
            mean            0
            covariance      2^(-1)
            decomp_type     PCA
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               1
            replications    25
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7
        
        Vector outputs
        
        >>> f = BoxIntegral(DigitalNetB2(3,seed=7,replications=25),s=[-1,1])
        >>> abs_tol = 1e-3
        >>> sc = CubQMCRepStudentT(f,abs_tol=abs_tol,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> solution
        array([1.19025707, 0.96062762])
        >>> data
        Data (Data)
            solution        [1.19  0.961]
            comb_bound_low  [1.19  0.961]
            comb_bound_high [1.191 0.961]
            comb_bound_diff [0.001 0.   ]
            comb_flags      [ True  True]
            n_total         409600
            n               [409600   6400]
            n_rep           [16384   256]
            time_integrate  ...
        CubQMCRepStudentT (AbstractStoppingCriterion)
            inflate         1
            alpha           0.010
            abs_tol         0.001
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(30)
        BoxIntegral (AbstractIntegrand)
            s               [-1  1]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               3
            replications    25
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7
        >>> sol3neg1 = -np.pi/4-1/2*np.log(2)+np.log(5+3*np.sqrt(3))
        >>> sol31 = np.sqrt(3)/4+1/2*np.log(2+np.sqrt(3))-np.pi/24
        >>> true_value = np.array([sol3neg1,sol31])
        >>> assert (abs(true_value-solution)<abs_tol).all()
        
        Sensitivity indices 

        >>> function = Genz(DigitalNetB2(3,seed=7,replications=25))
        >>> integrand = SensitivityIndices(function)
        >>> sc = CubQMCRepStudentT(integrand,abs_tol=5e-4,rel_tol=0)
        >>> solution,data = sc.integrate()
        >>> data
        Data (Data)
            solution        [[0.02  0.196 0.667]
                             [0.036 0.303 0.782]]
            comb_bound_low  [[0.019 0.195 0.667]
                             [0.035 0.303 0.781]]
            comb_bound_high [[0.02  0.196 0.667]
                             [0.036 0.303 0.782]]
            comb_bound_diff [[0.001 0.001 0.001]
                             [0.001 0.001 0.001]]
            comb_flags      [[ True  True  True]
                             [ True  True  True]]
            n_total         204800
            n               [[[102400 204800 204800]
                              [102400 204800 204800]
                              [102400 204800 204800]]
        <BLANKLINE>
                             [[ 12800 102400 102400]
                              [ 12800 102400 102400]
                              [ 12800 102400 102400]]]
            n_rep           [[[4096 8192 8192]
                              [4096 8192 8192]
                              [4096 8192 8192]]
        <BLANKLINE>
                             [[ 512 4096 4096]
                              [ 512 4096 4096]
                              [ 512 4096 4096]]]
            time_integrate  ...
        CubQMCRepStudentT (AbstractStoppingCriterion)
            inflate         1
            alpha           0.010
            abs_tol         5.00e-04
            rel_tol         0
            n_init          2^(8)
            n_limit         2^(30)
        SensitivityIndices (AbstractIntegrand)
            indices         [[ True False False]
                             [False  True False]
                             [False False  True]]
        Uniform (AbstractTrueMeasure)
            lower_bound     0
            upper_bound     1
        DigitalNetB2 (AbstractLDDiscreteDistribution)
            d               3
            replications    25
            randomize       LMS DS
            gen_mats_source joe_kuo.6.21201.txt
            order           RADICAL INVERSE
            t               63
            alpha           1
            n_limit         2^(32)
            entropy         7
        
    
    **References:**
    
    1.  Art B. Owen. "Practical Quasi-Monte Carlo Integration." 2023.  
        [https://artowen.su.domains/mc/](https://artowen.su.domains/mc/). 

    2.  Pierre l’Ecuyer et al.  
        "Confidence intervals for randomized quasi-Monte Carlo estimators."  
        2023 Winter Simulation Conference (WSC). IEEE, 2023.  
        [https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=10408613](https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=10408613). 
    """

    def __init__(self,
                 integrand, 
                 abs_tol = 1e-2,
                 rel_tol = 0.,
                 n_init = 256.,
                 n_limit = 2**30,
                 error_fun = "EITHER",
                 inflate = 1,
                 alpha = 0.01, 
                 ):
        r"""
        Args:
            integrand (AbstractIntegrand): The integrand.
            abs_tol (np.ndarray): Absolute error tolerance.
            rel_tol (np.ndarray): Relative error tolerance.
            n_init (int): Initial number of samples. 
            n_limit (int): Maximum number of samples.
            error_fun (Union[str,callable]): Function mapping the approximate solution, absolute error tolerance, and relative error tolerance to the current error bound.

                - `'EITHER'`, the default, requires the approximation error must be below either the absolue *or* relative tolerance.
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.maximum(abs_tol,abs(sv)*rel_tol)
                    ```
                - `'BOTH'` requires the approximation error to be below both the absolue *and* relative tolerance. 
                    Equivalent to setting
                    ```python
                    error_fun = lambda sv,abs_tol,rel_tol: np.minimum(abs_tol,abs(sv)*rel_tol)
                    ```
            inflate (float): Inflation factor $\geq 1$ to multiply by the variance estimate to make it more conservative.
            alpha (np.ndarray): Uncertainty level in $(0,1)$. 
        """
        self.parameters = ['inflate','alpha','abs_tol','rel_tol','n_init','n_limit']
        # Input Checks
        if np.log2(n_init)%1!=0:
            warnings.warn('n_init must be a power of two. Using n_init = 2**5',ParameterWarning)
            n_init = 2**5
        if np.log2(n_limit)%1!=0:
            warnings.warn('n_init must be a power of two. Using n_limit = 2**30',ParameterWarning)
            n_limit = 2**30
        # Set Attributes
        self.n_init = int(n_init)
        self.n_limit = int(n_limit)
        assert isinstance(error_fun,str) or callable(error_fun)
        if isinstance(error_fun,str):
            if error_fun.upper()=="EITHER":
                error_fun = lambda sv,abs_tol,rel_tol: np.maximum(abs_tol,abs(sv)*rel_tol)
            elif error_fun.upper()=="BOTH":
                error_fun = lambda sv,abs_tol,rel_tol: np.minimum(abs_tol,abs(sv)*rel_tol)
            else:
                raise ParameterError("str error_fun must be 'EITHER' or 'BOTH'")
        self.error_fun = error_fun
        self.alpha = alpha
        self.inflate = float(inflate)
        assert self.inflate>=1
        assert 0<self.alpha<1
        # QMCPy Objs
        self.integrand = integrand
        self.true_measure = self.integrand.true_measure
        self.discrete_distrib = self.integrand.discrete_distrib
        super(CubQMCRepStudentT,self).__init__(allowed_distribs=[AbstractLDDiscreteDistribution],allow_vectorized_integrals=True)
        assert self.integrand.discrete_distrib.replications>1, "Require the discrete distribution has replications>1"
        assert self.integrand.discrete_distrib.randomize!="FALSE", "Require discrete distribution is randomized"
        self.alphas_indv,identity_dependency = self._compute_indv_alphas(np.full(self.integrand.d_comb,self.alpha))
        self.set_tolerance(abs_tol,rel_tol)
        self.t_star = -t.ppf(self.alphas_indv/2,df=self.integrand.discrete_distrib.replications-1)
        
    def integrate(self):
        t_start = time()
        data = Data(
            parameters = [
                'solution',
                'comb_bound_low',
                'comb_bound_high',
                'comb_bound_diff',
                'comb_flags',
                'n_total',
                'n',
                'n_rep',
                'time_integrate'])
        data.flags_indv = np.tile(False,self.integrand.d_indv)
        data.compute_flags = np.tile(True,self.integrand.d_indv)
        data.n_rep = np.tile(self.n_init,self.integrand.d_indv)
        data.n_min = 0
        data.n_max = self.n_init
        data.solution_indv = np.tile(np.nan,self.integrand.d_indv)
        data.xfull = np.empty((self.discrete_distrib.replications,0,self.integrand.d))
        data.yfull = np.empty(self.integrand.d_indv+(self.discrete_distrib.replications,0))
        data._ysums = np.zeros(self.integrand.d_indv+(self.discrete_distrib.replications,),dtype=float)
        while True:
            xnext = self.discrete_distrib(n_min=data.n_min,n_max=data.n_max)
            data.xfull = np.concatenate([data.xfull,xnext],1)
            ynext = self.integrand.f(xnext,compute_flags=data.compute_flags)
            ynext[~data.compute_flags] = np.nan
            data.yfull = np.concatenate([data.yfull,ynext],-1)
            data.n_rep[data.compute_flags] = data.n_max
            data._ysums[data.compute_flags] += ynext[data.compute_flags].sum(-1)
            data.muhats = data._ysums/data.n_rep[...,None]
            data.solution_indv = data.muhats.mean(-1)
            data.sigmahat = data.muhats.std(-1,ddof=1)
            data.ci_half_width = self.t_star*self.inflate*data.sigmahat/np.sqrt(self.discrete_distrib.replications)
            data.indv_bound_low = data.solution_indv-data.ci_half_width
            data.indv_bound_high = data.solution_indv+data.ci_half_width
            data.n = self.discrete_distrib.replications*data.n_rep
            data.n_total = data.n.max() 
            data.comb_bound_low,data.comb_bound_high = self.integrand.bound_fun(data.indv_bound_low,data.indv_bound_high)
            data.comb_bound_diff = data.comb_bound_high-data.comb_bound_low
            fidxs = np.isfinite(data.comb_bound_low)&np.isfinite(data.comb_bound_high)
            slow,shigh,abs_tols,rel_tols = data.comb_bound_low[fidxs],data.comb_bound_high[fidxs],self.abs_tols[fidxs],self.rel_tols[fidxs]
            data.solution = np.tile(np.nan,data.comb_bound_low.shape)
            data.solution[fidxs] = 1/2*(slow+shigh+self.error_fun(slow,abs_tols,rel_tols)-self.error_fun(shigh,abs_tols,rel_tols))
            data.comb_flags = np.tile(False,data.comb_bound_low.shape)
            data.comb_flags[fidxs] = (shigh-slow) <= (self.error_fun(slow,abs_tols,rel_tols)+self.error_fun(shigh,abs_tols,rel_tols))
            data.flags_indv = self.integrand.dependency(data.comb_flags)
            data.compute_flags = ~data.flags_indv
            if np.sum(data.compute_flags)==0:
                break # sufficiently estimated
            elif 2*data.n_total>self.n_limit:
                warning_s = """
                Already generated %d samples.
                Trying to generate %d new samples would exceeds n_limit = %d.
                No more samples will be generated.
                Note that error tolerances may not be satisfied. """ \
                % (int(data.n_total),int(data.n_total),int(self.n_limit))
                warnings.warn(warning_s, MaxSamplesWarning)
                break
            data.n_min = data.n_max
            data.n_max = 2*data.n_min
        data.stopping_crit = self
        data.integrand = self.integrand
        data.true_measure = self.integrand.true_measure
        data.discrete_distrib = self.true_measure.discrete_distrib
        data.time_integrate = time()-t_start
        return data.solution,data
    
    def set_tolerance(self, abs_tol=None, rel_tol=None, rmse_tol=None):
        assert rmse_tol is None, "rmse_tol not supported by this stopping criterion."
        if abs_tol is not None:
            self.abs_tol = abs_tol
            self.abs_tols = np.full(self.integrand.d_comb,self.abs_tol)
        if rel_tol is not None:
            self.rel_tol = rel_tol
            self.rel_tols = np.full(self.integrand.d_comb,self.rel_tol)




from .abstract_integrand import AbstractIntegrand
from ..true_measure import BrownianMotion
from ..discrete_distribution import DigitalNetB2
from ..util import ParameterError
import numpy as np
from scipy.stats import norm 


class FinancialOption(AbstractIntegrand):
    r"""
    Financial options.
    
    - Start price $S_0$
    - Strike price $K$
    - Interest rate $r$ 
    - Volatility $\sigma$
    - Drift $\gamma$
    - Equidistant monitoring times $\boldsymbol{\tau} = (\tau_1,\dots,\tau_d)^T$ with $\tau_d$ the final (exercise) time and $\tau_j = \tau_d j/d$. 

    Define the [geometric brownian motion](https://en.wikipedia.org/wiki/Geometric_Brownian_motion) as 

    $$\boldsymbol{S}(\boldsymbol{t}) = S_0 e^{(\gamma-\sigma^2/2)\boldsymbol{\tau}+\sigma\boldsymbol{t}}, \qquad \boldsymbol{T} \sim \mathcal{N}(\boldsymbol{0},\mathsf{\Sigma})$$

    where $\boldsymbol{T}$ is a standard Brownian motion so $\mathsf{\Sigma} = \left(\min\{\tau_j,\tau_{j'}\}\right)_{j,j'=1}^d$.

    The discounted payoff is 

    $$g(\boldsymbol{t}) = P(\boldsymbol{S}(\boldsymbol{t}))e^{-r \tau_d}$$
    
    where the payoff function $P$ will be defined depending on the option.
    
    Below we wil use $S_{-1}$ to denote the final element of $\boldsymbol{S}$, the value of the path at exercise time. 
    
    # European Options
    
    *European Call and Put Options* have respective payoffs

    $$P(\boldsymbol{S}) = \max\{S_{-1}-K,0\}, \qquad P(\boldsymbol{S}) = \max\{K-S_{-1},0\}.$$

    # Asian Options

    An asian option considers the average value of an asset path across time. We use the trapezoidal rule to approximate either the *arithmetic mean* by 

    $$A(\boldsymbol{S}) = \frac{1}{d}\left[\frac{1}{2} S_0 + \sum_{j=1}^{d-1} S_j + \frac{1}{2} S_{-1}\right]$$ 
    
    or the *geometric mean* by 

    $$A(\boldsymbol{S}) = \left[\sqrt{S_0} \prod_{j=1}^{d-1} S_j \sqrt{S_{-1}}\right]^{1/d}.$$

    *Asian Call and Put Option* have respective payoffs 

    $$P(\boldsymbol{S}) = \max\{A(\boldsymbol{S})-K,0\}, \qquad P(\boldsymbol{S}) = \max\{K-A(\boldsymbol{S}),0\}.$$

    # Barrier Options 

    - Barrier $B$. 
    
    *In* options are activate when the path crosses the barrier $B$, while *out* options are activated only if the path never crosses the barrier $B$.
    An *up* option satisfies $S_0<B$ while a *down* option satisfies $S_0>B$, both indicating the direction of the barrier from the start price. 

    *Barrier Up-In Call and Put Options* have respective payoffs

    $$P(\boldsymbol{S}) = \begin{cases} \max\{S_{-1})-K,0\}, & \text{any } \boldsymbol{S} \geq B \\ 0, & \mathrm{otherwise} \end{cases}, \qquad P(\boldsymbol{S}) = \begin{cases} \max\{K-S_{-1}),0\}, & \text{any } \boldsymbol{S} \geq B \\ 0, & \mathrm{otherwise} \end{cases}.$$

    *Barrier Up-Out Call and Put Options* have respective payoffs 

    $$P(\boldsymbol{S}) = \begin{cases} \max\{S_{-1})-K,0\}, & \text{all } \boldsymbol{S} < B \\ 0, & \mathrm{otherwise} \end{cases}, \qquad P(\boldsymbol{S}) = \begin{cases} \max\{K-S_{-1}),0\}, & \text{all } \boldsymbol{S} < B \\ 0, & \mathrm{otherwise} \end{cases}.$$
    
    *Barrier Down-In Call and Put Options* have respective payoffs 

    $$P(\boldsymbol{S}) = \begin{cases} \max\{S_{-1})-K,0\}, & \text{any } \boldsymbol{S} \leq B \\ 0, & \mathrm{otherwise} \end{cases}, \qquad P(\boldsymbol{S}) = \begin{cases} \max\{K-S_{-1}),0\}, & \text{any } \boldsymbol{S} \leq B \\ 0, & \mathrm{otherwise} \end{cases}.$$

    *Barrier Down-Out Call and Put Options* have respective payoffs 

    $$P(\boldsymbol{S}) = \begin{cases} \max\{S_{-1})-K,0\}, & \text{all } \boldsymbol{S} > B \\ 0, & \mathrm{otherwise} \end{cases}, \qquad P(\boldsymbol{S}) = \begin{cases} \max\{K-S_{-1}),0\}, & \text{all } \boldsymbol{S} > B \\ 0, & \mathrm{otherwise} \end{cases}.$$

    # Lookback Options 

    *Lookback Call and Put Options* have respective payoffs

    $$P(\boldsymbol{S}) = S_{-1}-\min(\boldsymbol{S}), \qquad P(\boldsymbol{S}) = \max(\boldsymbol{S})-S_{-1}.$$
    
    # Digital Option 

    - Payout $\rho$. 

    *Digital Call and Put Options* have respective payoffs

    $$P(\boldsymbol{S}) = \begin{cases} \rho, & S_{-1} \geq K \\ 0, & \mathrm{otherwise} \end{cases}, \qquad P(\boldsymbol{S}) =  \begin{cases} \rho, & S_{-1} \leq K \\ 0, & \mathrm{otherwise} \end{cases}.$$

    # Multilevel Options

    - Initial level $\ell_0 \geq 0$. 
    - Level $\ell \geq \ell_0$.
    
    Let $\boldsymbol{S}_\mathrm{fine}=\boldsymbol{S}$ be the *fine* full path. For $\ell>\ell_0$ write the *coarse* path as $\boldsymbol{S}_\mathrm{coarse} = (S_j)_{j \text{ even}}$ which only considers every other element of $\boldsymbol{S}$. 
    In this multilevel setting the payoff is

    $$P_\ell(\boldsymbol{S}) = \begin{cases} P(\boldsymbol{S}_\mathrm{fine}), & \ell = \ell_0, \\ P(\boldsymbol{S}_\mathrm{fine})-P(\boldsymbol{S}_\mathrm{coarse}), & \ell > \ell_0 \end{cases}.$$

    Cancellations from the telescoping sum allow us to write 

    $$\lim_{\ell \to \infty} P_\ell = P_{\ell_0} + \sum_{\ell=\ell_0+1}^\infty P_\ell.$$

    Examples:
        >>> integrand = FinancialOption(DigitalNetB2(dimension=3,seed=7),option="EUROPEAN")
        >>> y = integrand(2**10)
        >>> y.shape
        (1024,)
        >>> print("%.4f"%y.mean())
        4.2126
        >>> print("%.4f"%integrand.get_exact_value())
        4.2115
        >>> integrand
        FinancialOption (AbstractIntegrand)
            option          EUROPEAN
            call_put        CALL
            volatility      2^(-1)
            start_price     30
            strike_price    35
            interest_rate   0
            t_final         1
        >>> integrand.true_measure
        BrownianMotion (AbstractTrueMeasure)
            time_vec        [0.333 0.667 1.   ]
            drift           0
            mean            [0. 0. 0.]
            covariance      [[0.333 0.333 0.333]
                             [0.333 0.667 0.667]
                             [0.333 0.667 1.   ]]
            decomp_type     PCA

        >>> integrand = FinancialOption(DigitalNetB2(dimension=64,seed=7),option="ASIAN")
        >>> y = integrand(2**10)
        >>> y.shape
        (1024,)
        >>> print("%.4f"%y.mean())
        1.7782

        With independent replications

        >>> integrand = FinancialOption(DigitalNetB2(dimension=64,seed=7,replications=2**4),option="ASIAN")
        >>> y = integrand(2**6)
        >>> y.shape
        (16, 64)
        >>> muhats = y.mean(-1) 
        >>> muhats.shape 
        (16,)
        >>> print("%.4f"%muhats.mean())
        1.7765

        Multi-level options 

        >>> seed_seq = np.random.SeedSequence(7) 
        >>> d_coarsest = 8
        >>> num_levels = 4
        >>> ns = [2**11,2**10,2**9,2**8]
        >>> integrands = [FinancialOption(DigitalNetB2(dimension=2**l*d_coarsest,seed=seed_seq.spawn(1)[0]),option="ASIAN",level=l,d_coarsest=d_coarsest) for l in range(num_levels)]
        >>> ys = [integrands[l](ns[l]) for l in range(num_levels)]
        >>> for l in range(num_levels):
        ...     print("ys[%d].shape = %s"%(l,ys[l].shape))
        ys[0].shape = (2, 2048)
        ys[1].shape = (2, 1024)
        ys[2].shape = (2, 512)
        ys[3].shape = (2, 256)
        >>> ymeans = np.stack([(ys[l][1]-ys[l][0]).mean(-1) for l in range(num_levels)])
        >>> with np.printoptions(formatter={"float": lambda x: "%.2e"%x}):
        ...     ymeans
        array([1.78e+00, 2.61e-03, 5.57e-04, 1.10e-03])
        >>> print("%.4f"%ymeans.sum())
        1.7850

        Multi-level options with independent replications
         
        >>> seed_seq = np.random.SeedSequence(7) 
        >>> d_coarsest = 8
        >>> num_levels = 4
        >>> ns = [2**7,2**6,2**5,2**4]
        >>> integrands = [FinancialOption(DigitalNetB2(dimension=2**l*d_coarsest,seed=seed_seq.spawn(1)[0],replications=2**4),option="ASIAN",level=l,d_coarsest=d_coarsest) for l in range(num_levels)]
        >>> ys = [integrands[l](ns[l]) for l in range(num_levels)]
        >>> for l in range(num_levels):
        ...     print("ys[%d].shape = %s"%(l,ys[l].shape))
        ys[0].shape = (2, 16, 128)
        ys[1].shape = (2, 16, 64)
        ys[2].shape = (2, 16, 32)
        ys[3].shape = (2, 16, 16)
        >>> muhats = np.stack([(ys[l][1]-ys[l][0]).mean(-1) for l in range(num_levels)])
        >>> muhats.shape
        (4, 16)
        >>> muhathat = muhats.mean(-1)
        >>> with np.printoptions(formatter={"float": lambda x: "%.2e"%x}):
        ...     muhathat
        array([1.80e+00, -3.08e-03, 2.64e-03, 1.14e-03])
        >>> print("%.4f"%muhathat.sum())
        1.7982

    **References:**

    1.  M.B. Giles.  
        Improved multilevel Monte Carlo convergence using the Milstein scheme.  
        343-358, in Monte Carlo and Quasi-Monte Carlo Methods 2006, Springer, 2008.  
        [http://people.maths.ox.ac.uk/~gilesm/files/mcqmc06.pdf](http://people.maths.ox.ac.uk/~gilesm/files/mcqmc06.pdf).
    """
                          
    def __init__(self, 
                 sampler,
                 option = "ASIAN", 
                 call_put = 'CALL',
                 volatility = 0.5,
                 start_price = 30,
                 strike_price = 35,
                 interest_rate = 0,
                 t_final = 1,
                 decomp_type = 'PCA',
                 level = None,
                 d_coarsest = 2,
                 asian_mean = "ARITHMETIC",
                 asian_mean_quadrature_rule = "TRAPEZOIDAL",
                 barrier_in_out = "IN",
                 barrier_price = 38,
                 digital_payout = 10):
        r"""
        Args:
            sampler (Union[AbstractDiscreteDistribution,AbstractTrueMeasure]): Either  
                
                - a discrete distribution from which to transform samples, or
                - a true measure by which to compose a transform.
            option (str): Option type in `['ASIAN', 'EUROPEAN', 'BARRIER', 'LOOKBACK', 'DIGITAL']`
            call_put (str): Either `'CALL'` or `'PUT'`. 
            volatility (float): $\sigma$.
            start_price (float): $S_0$.
            strike_price (float): $K$.
            interest_rate (float): $r$.
            t_final (float): $\tau_d$.
            decomp_type (str): Method for decomposition for covariance matrix. Options include
             
                - `'PCA'` for principal component analysis, or 
                - `'Cholesky'` for cholesky decomposition.
            level (Union[None,int]): Level for multilevel problems 
            d_coarsest (Union[None,int]): Dimension of the problem on the coarsest level.
            asian_mean (str): Either `'ARITHMETIC'` or `'GEOMETRIC'`.
            asian_mean_quadrature_rule (str): Either 'TRAPEZOIDAL' or 'RIGHT'. 
            barrier_in_out (str): Either `'IN'` or `'OUT'`. 
            barrier_price (float): $B$. 
            digital_payout (float): $\rho$. 
        """
        self.parameters = ['option', 'call_put', 'volatility', 'start_price', 'strike_price', 'interest_rate', 't_final']
        self.t_final = t_final
        self.sampler = sampler
        self.decomp_type = decomp_type
        self.true_measure = BrownianMotion(self.sampler,t_final=self.t_final,decomp_type=self.decomp_type)
        self.volatility = float(volatility)
        self.start_price = float(start_price)
        self.strike_price = float(strike_price)
        self.interest_rate = float(interest_rate)
        self.discount_factor = np.exp(-self.interest_rate*self.t_final)
        self.level = level
        self.d_coarsest = d_coarsest
        if self.level is not None:
            self.multilevel = True 
            self.parameters += ['level','d_coarsest']
            assert np.isscalar(self.level) and self.level%1==0 
            assert np.isscalar(self.d_coarsest) and self.d_coarsest%1==0 and d_coarsest>0 and np.log2(d_coarsest)%1==0, "d_coarsest must be an integer power of 2"
            self.level = int(self.level)
            self.d_coarsest = int(self.d_coarsest) 
            assert self.sampler.d==self.d_coarsest*2**self.level, "the dimension of the sampler must equal d_coarsest*2^level = %d"%(d_coarsest*2**self.level)
            self.cost = self.d_coarsest*2**self.level
            dim_shape = (2,)
        else:
            self.multilevel = False
            dim_shape = ()
        self.call_put = str(call_put).upper()
        assert self.call_put in ['CALL','PUT'], "invalid call_put = %s"%self.call_put
        self.option = str(option).upper()
        self.asian_mean = str(asian_mean).upper()
        self.asian_mean_quadrature_rule = str(asian_mean_quadrature_rule).upper()
        self.barrier_in_out = str(barrier_in_out).upper()
        assert np.isscalar(barrier_price)
        self.barrier_price = float(barrier_price)
        assert np.isscalar(digital_payout) and digital_payout>0
        self.digital_payout = float(digital_payout)
        if self.option=="EUROPEAN":
            self.payoff = self.payoff_european_call if self.call_put=='CALL' else self.payoff_european_put
        elif self.option=="ASIAN":
            self.parameters += ['asian_mean']
            assert self.asian_mean in ['ARITHMETIC','GEOMETRIC'], "invalid asian_mean = %s"%self.asian_mean
            assert self.asian_mean_quadrature_rule in ['TRAPEZOIDAL','RIGHT'], "invalid asian_mean_quadrature_rule = %s"%self.asian_mean_quadrature_rule
            if self.asian_mean=="ARITHMETIC":
                if self.asian_mean_quadrature_rule=="TRAPEZOIDAL":
                    self.payoff = self.payoff_asian_arithmetic_trap_call if self.call_put=='CALL' else self.payoff_asian_arithmetic_trap_put
                elif self.asian_mean_quadrature_rule=="RIGHT":
                    self.payoff = self.payoff_asian_arithmetic_right_call if self.call_put=='CALL' else self.payoff_asian_arithmetic_right_put
            elif self.asian_mean=="GEOMETRIC":
                if self.asian_mean_quadrature_rule=="TRAPEZOIDAL":
                    self.payoff = self.payoff_asian_geometric_trap_call if self.call_put=='CALL' else self.payoff_asian_geometric_trap_put
                elif self.asian_mean_quadrature_rule=="RIGHT":
                    self.payoff = self.payoff_asian_geometric_right_call if self.call_put=='CALL' else self.payoff_asian_geometric_right_put
        elif self.option=="BARRIER":
            self.parameters += ['barrier_in_out','barrier_up_down','barrier_price']
            self.barrier_up_down = "UP" if self.start_price<self.barrier_price else "DOWN"
            if self.barrier_in_out=="IN":
                if self.barrier_up_down=="UP":
                    self.payoff = self.payoff_barrier_in_up_call if self.call_put=='CALL' else self.payoff_barrier_in_up_put
                else:
                    self.payoff = self.payoff_barrier_in_down_call if self.call_put=='CALL' else self.payoff_barrier_in_down_put
            elif self.barrier_in_out=="OUT":
                if self.barrier_up_down=="UP":
                    self.payoff = self.payoff_barrier_out_up_call if self.call_put=='CALL' else self.payoff_barrier_out_up_put
                else:
                    self.payoff = self.payoff_barrier_out_down_call if self.call_put=='CALL' else self.payoff_barrier_out_down_put
            else:
                raise ParameterError("invalid barrier_in_out = %s"%self.barrier_in_out)
        elif self.option=="LOOKBACK":
            self.payoff = self.payoff_lookback_call if self.call_put=='CALL' else self.payoff_lookback_put
        elif self.option=="DIGITAL":
            self.payoff = self.payoff_digital_call if self.call_put=='CALL' else self.payoff_digital_put
        else:
            raise ParameterError("invalid option type %s"%self.option)
        super(FinancialOption,self).__init__(dimension_indv=dim_shape,dimension_comb=dim_shape,parallel=False)  
    
    def g(self, t, **kwargs):
        gbm = self.gbm(t)
        discounted_payoffs = self.payoff(gbm)*self.discount_factor
        if self.multilevel:
            if self.level==0:
                discounted_payoffs_coarse = np.zeros_like(discounted_payoffs)
            else: 
                discounted_payoffs_coarse = self.payoff(gbm[...,1::2])*self.discount_factor
            discounted_payoffs = np.stack([discounted_payoffs_coarse,discounted_payoffs])
        return discounted_payoffs
    
    def gbm(self, t):
        gbm = self.start_price*np.exp((self.interest_rate-self.volatility**2/2)*self.true_measure.time_vec+self.volatility*t)
        gbm *= (gbm>0).cumprod(-1) # if a path hits 0, set remaining values in the path to 0
        return gbm
    
    def payoff_european_call(self, gbm):
        return np.maximum(gbm[...,-1]-self.strike_price,0)
    
    def payoff_european_put(self, gbm):
        return np.maximum(self.strike_price-gbm[...,-1],0)
    
    def payoff_asian_arithmetic_trap_call(self, gbm):
        return np.maximum((self.start_price/2+gbm[...,:-1].sum(-1)+gbm[...,-1]/2)/gbm.shape[-1]-self.strike_price,0)
    
    def payoff_asian_arithmetic_trap_put(self, gbm):
        return np.maximum(self.strike_price-(self.start_price/2+gbm[...,:-1].sum(-1)+gbm[...,-1]/2)/gbm.shape[-1],0)
    
    def payoff_asian_geometric_trap_call(self, gbm):
        return np.maximum(np.exp((np.log(self.start_price)/2+np.log(gbm[...,:-1]).sum(-1)+np.log(gbm[...,-1])/2)/gbm.shape[-1])-self.strike_price,0)
    
    def payoff_asian_geometric_trap_put(self, gbm):
        return np.maximum(self.strike_price-np.exp((np.log(self.start_price)/2+np.log(gbm[...,:-1]).sum(-1)+np.log(gbm[...,-1])/2)/gbm.shape[-1]),0)
    
    def payoff_asian_arithmetic_right_call(self, gbm):
        return np.maximum(gbm.sum(-1)/gbm.shape[-1]-self.strike_price,0)
    
    def payoff_asian_arithmetic_right_put(self, gbm):
        return np.maximum((self.strike_price-gbm.sum(-1))/gbm.shape[-1],0)
    
    def payoff_asian_geometric_right_call(self, gbm):
        return np.maximum(np.exp(np.log(gbm).sum(-1)/gbm.shape[-1])-self.strike_price,0)
    
    def payoff_asian_geometric_right_put(self, gbm):
        return np.maximum(self.strike_price-np.exp(np.log(gbm).sum(-1)/gbm.shape[-1]),0)
    
    def payoff_barrier_in_up_call(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm>=self.barrier_price).any(-1)
        v[~flag] = 0
        v[flag] = np.maximum(v[flag]-self.strike_price,0)
        return v
    
    def payoff_barrier_out_up_call(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm<self.barrier_price).all(-1)
        v[~flag] = 0
        v[flag] = np.maximum(v[flag]-self.strike_price,0)
        return v
    
    def payoff_barrier_in_down_call(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm<=self.barrier_price).any(-1)
        v[~flag] = 0
        v[flag] = np.maximum(v[flag]-self.strike_price,0)
        return v
    
    def payoff_barrier_out_down_call(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm>self.barrier_price).all(-1)
        v[~flag] = 0
        v[flag] = np.maximum(v[flag]-self.strike_price,0)
        return v
    
    def payoff_barrier_in_up_put(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm>=self.barrier_price).any(-1)
        v[~flag] = 0
        v[flag] = np.maximum(self.strike_price-v[flag],0)
        return v
    
    def payoff_barrier_out_up_put(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm<self.barrier_price).all(-1)
        v[~flag] = 0
        v[flag] = np.maximum(self.strike_price-v[flag],0)
        return v
    
    def payoff_barrier_in_down_put(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm<=self.barrier_price).any(-1)
        v[~flag] = 0
        v[flag] = np.maximum(self.strike_price-v[flag],0)
        return v
    
    def payoff_barrier_out_down_put(self, gbm):
        v = gbm[...,-1].copy()
        flag = (gbm>self.barrier_price).all(-1)
        v[~flag] = 0
        v[flag] = np.maximum(self.strike_price-v[flag],0)
        return v
    
    def payoff_lookback_call(self, gbm):
        return gbm[...,-1]-gbm.min(-1)
    
    def payoff_lookback_put(self, gbm):
        return gbm.max(-1)-gbm[...,-1]
    
    def payoff_digital_call(self, gbm):
        return np.where(gbm[...,-1]>=self.strike_price,self.digital_payout,0)
    
    def payoff_digital_put(self, gbm):
        return np.where(gbm[...,-1]<=self.strike_price,self.digital_payout,0)
    
    def get_exact_value(self):
        """
        Compute the exact analytic fair price of the option in finite dimensions. Supports 
            
        - `option='EUROPEAN'`
        - `option='ASIAN'` with `asian_mean='GEOMETRIC'` and `asian_mean_quadrature_rule='RIGHT'`

        Returns: 
            mean (float): Exact value of the integral. 
        """
        if self.option=="EUROPEAN":
            denom = self.volatility*np.sqrt(self.t_final)
            decay = self.strike_price*self.discount_factor
            if self.call_put == 'CALL':
                term1 = np.log(self.start_price/self.strike_price)+(self.interest_rate+self.volatility**2/2)*self.t_final
                term2 = np.log(self.start_price/self.strike_price)+(self.interest_rate-self.volatility**2/2)*self.t_final
                fp = self.start_price*norm.cdf(term1/denom)-decay*norm.cdf(term2/denom)
            elif self.call_put == 'PUT':
                term1 = np.log(self.strike_price/self.start_price)-(self.interest_rate-self.volatility**2/2)*self.t_final
                term2 = np.log(self.strike_price/self.start_price)-(self.interest_rate+self.volatility**2/2)*self.t_final
                fp = decay*norm.cdf(term1/denom)-self.start_price*norm.cdf(term2/denom)
        elif self.option=="ASIAN":
            assert self.asian_mean=='GEOMETRIC' and self.asian_mean_quadrature_rule=='RIGHT', "exact value for Asian options only implemented for self.asian_mean=='GEOMETRIC' and self.asian_mean_quadrature_rule=='RIGHT'"
            Tbar = (1+1/self.d)*self.t_final/2
            sigmabar = self.volatility*np.sqrt((2+1/self.d)/3)
            rbar = self.interest_rate+(sigmabar**2-self.volatility**2)/2
            gmeancall,gmeanput = _eurogbmprice(self.start_price,rbar,Tbar,sigmabar,self.strike_price)
            if self.call_put=='CALL':
                fp = gmeancall * np.exp(rbar*Tbar-self.interest_rate*self.t_final)
            elif self.call_put=='PUT':
                fp = gmeanput * np.exp(rbar*Tbar-self.interest_rate*self.t_final)
            return fp
        else:
            raise ParameterError("exact value not supported for option = %s"%self.option)
        return fp
    
    def get_exact_value_inf_dim(self):
        r"""
        Get the exact analytic fair price of the option in infinite dimensions. Supports 

        - `option='ASIAN'` with `asian_mean='GEOMETRIC'`

        Returns: 
            mean (float): Exact value of the integral. 
        """
        if self.option=='ASIAN':
            assert self.asian_mean=='GEOMETRIC' , "get_exact_value_inf_dim for the Asian option only available for self.asian_mean=='GEOMETRIC'"
            sigma_g = self.volatility/np.sqrt(3) 
            b = 1/2*(self.interest_rate-1/2*sigma_g**2)
            d1 = (np.log(self.start_price/self.strike_price)+(b+1/2*sigma_g**2)*self.t_final)/(sigma_g*np.sqrt(self.t_final))
            d2 = d1-sigma_g*np.sqrt(self.t_final)
            f1 = self.start_price*np.exp((b-self.interest_rate)*self.t_final)
            f2 = self.strike_price*np.exp(-self.interest_rate*self.t_final)
            if self.call_put=="CALL":
                val = f1*norm.cdf(d1)-f2*norm.cdf(d2)
            elif self.call_put=="PUT":
                val = f2*norm.cdf(-d2)-f1*norm.cdf(-d1)
        else:
            raise Exception("get_exact_value_inf_dim not implemented for option = %s"%self.option)
        return val
    
    def dimension_at_level(self, level):
        return self.d_coarsest*2**level
    
    def _spawn(self, level, sampler):
        return FinancialOption(
            sampler=sampler,
            option = self.option,
            call_put = self.call_put,
            volatility=self.volatility,
            start_price=self.start_price,
            strike_price=self.strike_price,
            interest_rate=self.interest_rate,
            t_final=self.t_final,
            decomp_type = self.decomp_type,
            level=level, 
            d_coarsest=self.d_coarsest,
            asian_mean=self.asian_mean,
            asian_mean_quadrature_rule = self.asian_mean_quadrature_rule,
            barrier_in_out=self.barrier_in_out, 
            barrier_price=self.barrier_price,
            digital_payout=self.digital_payout)

def _eurogbmprice(S0, r, T, sigma, K):
    priceratio = K*np.exp(-r*T)/S0
    xbig = np.log(priceratio)/(sigma*np.sqrt(T))+sigma*np.sqrt(T)/2
    xsmall = np.log(priceratio)/(sigma*np.sqrt(T))-sigma*np.sqrt(T)/2
    putprice = S0*(priceratio*norm.cdf(xbig)-norm.cdf(xsmall))
    callprice = putprice+S0*(1-priceratio)
    return callprice,putprice

class AsianOption(FinancialOption):
    def __init__(self, *args, **kwargs):
        """ Deprecated, please use FinancialOption"""
        if "option" in kwargs:
            raise ParameterError("please do not pass 'option' to AsianOption")
        super(AsianOption,self).__init__(*args, **kwargs, option="ASIAN")

class EuropeanOption(FinancialOption):
    def __init__(self, *args, **kwargs):
        """ Deprecated, please use FinancialOption"""
        if "option" in kwargs:
            raise ParameterError("please do not pass 'option' to EuropeanOption")
        super(EuropeanOption,self).__init__(*args, **kwargs, option="EUROPEAN")

class BarrierOption(FinancialOption):
    def __init__(self, *args, **kwargs):
        """ Deprecated, please use FinancialOption"""
        if "option" in kwargs:
            raise ParameterError("please do not pass 'option' to BarrierOption")
        super(BarrierOption,self).__init__(*args, **kwargs, option="BARRIER")

class LookbackOption(FinancialOption):
    def __init__(self, *args, **kwargs):
        """ Deprecated, please use FinancialOption"""
        if "option" in kwargs:
            raise ParameterError("please do not pass 'option' to LookbackOption")
        super(LookbackOption,self).__init__(*args, **kwargs, option="LOOKBACK")

class DigitalOption(FinancialOption):
    def __init__(self, *args, **kwargs):
        """ Deprecated, please use FinancialOption"""
        if "option" in kwargs:
            raise ParameterError("please do not pass 'option' to DigitalOption")
        super(DigitalOption,self).__init__(*args, **kwargs, option="DIGITAL")



from .abstract_kernel import AbstractKernelScaleLengthscales
from ..discrete_distribution import DigitalNetB2
from ..util.transforms import tf_exp_eps,tf_exp_eps_inv,tf_identity
from ..util import ParameterError
import numpy as np 
import scipy.stats
import scipy.special


class AbstractKernelGaussianSE(AbstractKernelScaleLengthscales):

    AUTOGRADKERNEL = True 
    
    def parsed_single_integral_01d(self, x, batch_params):
        s = batch_params["scale"][...,0]
        l = batch_params["lengthscales"]
        nls = l.shape[-1]
        norm_class = self.npt.distributions.Normal if self.torchify else scipy.stats.norm
        norm = norm_class(x,l)
        lb = self.nptarray([0],**self.nptkwargs)
        ub = self.nptarray([1],**self.nptkwargs)
        kint = s*self.npt.prod(np.sqrt(2*np.pi)*l*(norm.cdf(ub)-norm.cdf(lb)),-1)
        return kint
    
    def double_integral_01d(self):
        erf = self.npt.erf if self.torchify else scipy.special.erf
        s = self.scale[...,0]
        l = self.lengthscales
        nls = l.shape[-1]
        kintint = s*self.npt.prod(2*(-1+self.npt.exp(-1/(2*l**2)))*l**2+np.sqrt(2*np.pi)*l*erf(1/(np.sqrt(2)*l)),-1)**(self.d/nls)
        return kintint


class KernelGaussian(AbstractKernelGaussianSE):
    
    r"""
    Gaussian / Squared Exponential kernel implemented using the product of exponentials. 

    $$K(\boldsymbol{x},\boldsymbol{z}) = S \prod_{j=1}^d \exp\left(-\left(\frac{x_j-z_j}{\sqrt{2} \gamma_j}\right)^2\right)$$

    Examples:
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelGaussian(d=2)
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        array([1., 1., 1., 1.])
        >>> kernel(x[:,None,:],x[None,:,:])
        array([[1.        , 0.78888466, 0.9483142 , 0.8228498 ],
               [0.78888466, 1.        , 0.72380317, 0.62226176],
               [0.9483142 , 0.72380317, 1.        , 0.95613874],
               [0.8228498 , 0.62226176, 0.95613874, 1.        ]])
        
        Multiple randomizations 
        
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (6, 6, 5, 5)

        Batch hyperparameters 

        >>> kernel = KernelGaussian(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1])
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        Integrals 

        >>> kernel = KernelGaussian(
        ...     d = 2,
        ...     scale = rng.uniform(low=0,high=1,size=(3,1,)),
        ...     lengthscales = rng.uniform(low=0,high=1,size=(1,2)))
        >>> kintint = kernel.double_integral_01d()
        >>> kintint
        array([0.50079567, 0.11125229, 0.34760005])
        >>> x_qmc_4d = DigitalNetB2(4,seed=7)(2**16)
        >>> kintint_qmc = kernel(x_qmc_4d[:,:2],x_qmc_4d[:,2:]).mean(1)
        >>> kintint_qmc
        array([0.5007959 , 0.11125234, 0.34760021])
        >>> with np.printoptions(formatter={"float":lambda x: "%.1e"%x}):
        ...     np.abs(kintint-kintint_qmc)
        array([2.3e-07, 5.1e-08, 1.6e-07])
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kint = kernel.single_integral_01d(x)
        >>> kint
        array([[0.54610372, 0.4272801 , 0.43001936, 0.44688778],
               [0.12131753, 0.09492073, 0.09552926, 0.0992766 ],
               [0.37904817, 0.29657323, 0.29847453, 0.31018283]])
        >>> x_qmc_2d = DigitalNetB2(2,seed=7)(2**16)
        >>> kint_qmc = kernel(x[:,None,:],x_qmc_2d).mean(-1)
        >>> kint_qmc
        array([[0.54610372, 0.4272801 , 0.43001936, 0.44688778],
               [0.12131753, 0.09492073, 0.09552926, 0.0992766 ],
               [0.37904817, 0.29657323, 0.29847453, 0.31018283]])
        >>> with np.printoptions(formatter={"float":lambda x: "%.1e"%x}):
        ...     np.abs(kint-kint_qmc)
        array([[2.2e-13, 4.0e-13, 4.2e-12, 3.9e-12],
               [4.8e-14, 8.9e-14, 9.4e-13, 8.8e-13],
               [1.5e-13, 2.8e-13, 2.9e-12, 2.7e-12]])
        >>> k_1l = KernelGaussian(d=2,lengthscales=[.5])
        >>> k_2l = KernelGaussian(d=2,lengthscales=[.5,.5])
        >>> print("%.5f"%k_2l.double_integral_01d())
        0.58363
        >>> print("%.5f"%k_1l.double_integral_01d())
        0.58363
        >>> k_1l.single_integral_01d(x)
        array([0.58119655, 0.46559577, 0.57603494, 0.53494393])
        >>> k_2l.single_integral_01d(x)
        array([0.58119655, 0.46559577, 0.57603494, 0.53494393])
                
        PyTorch

        >>> import torch
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelGaussian(d=2,torchify=True)
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        tensor([1., 1., 1., 1.], dtype=torch.float64, grad_fn=<MulBackward0>)
        >>> kernel(x[:,None,:],x[None,:,:])
        tensor([[1.0000, 0.7889, 0.9483, 0.8228],
                [0.7889, 1.0000, 0.7238, 0.6223],
                [0.9483, 0.7238, 1.0000, 0.9561],
                [0.8228, 0.6223, 0.9561, 1.0000]], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Multiple randomizations 
        
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(6,5,2)))
        >>> kernel(x,x).shape 
        torch.Size([6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([6, 6, 5, 5])

        Batch hyperparameters 

        >>> kernel = KernelGaussian(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1],
        ...     torchify = True)
        >>> kernel(x,x).shape 
        torch.Size([4, 3, 6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([4, 3, 6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([4, 3, 6, 6, 5, 5])

        Integrals 

        >>> kernel = KernelGaussian(
        ...     d = 2,
        ...     torchify = True,
        ...     scale = torch.from_numpy(rng.uniform(low=0,high=1,size=(3,1,))),
        ...     lengthscales = torch.from_numpy(rng.uniform(low=0,high=1,size=(1,2))),
        ...     requires_grad_scale = False, 
        ...     requires_grad_lengthscales = False)
        >>> kintint = kernel.double_integral_01d()
        >>> kintint
        tensor([0.5008, 0.1113, 0.3476], dtype=torch.float64)
        >>> x_qmc_4d = torch.from_numpy(DigitalNetB2(4,seed=7)(2**16))
        >>> kintint_qmc = kernel(x_qmc_4d[:,:2],x_qmc_4d[:,2:]).mean(1)
        >>> kintint_qmc
        tensor([0.5008, 0.1113, 0.3476], dtype=torch.float64)
        >>> with np.printoptions(formatter={"float":lambda x: "%.1e"%x}):
        ...     torch.abs(kintint-kintint_qmc).numpy()
        array([2.3e-07, 5.1e-08, 1.6e-07])
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kint = kernel.single_integral_01d(x)
        >>> kint
        tensor([[0.5461, 0.4273, 0.4300, 0.4469],
                [0.1213, 0.0949, 0.0955, 0.0993],
                [0.3790, 0.2966, 0.2985, 0.3102]], dtype=torch.float64)
        >>> x_qmc_2d = torch.from_numpy(DigitalNetB2(2,seed=7)(2**16))
        >>> kint_qmc = kernel(x[:,None,:],x_qmc_2d).mean(-1)
        >>> kint_qmc
        tensor([[0.5461, 0.4273, 0.4300, 0.4469],
                [0.1213, 0.0949, 0.0955, 0.0993],
                [0.3790, 0.2966, 0.2985, 0.3102]], dtype=torch.float64)
        >>> with np.printoptions(formatter={"float":lambda x: "%.1e"%x}):
        ...     torch.abs(kint-kint_qmc).numpy()
        array([[2.2e-13, 4.0e-13, 4.2e-12, 3.9e-12],
               [4.8e-14, 8.9e-14, 9.4e-13, 8.8e-13],
               [1.5e-13, 2.8e-13, 2.9e-12, 2.7e-12]])
        >>> k_1l = KernelGaussian(d=2,lengthscales=[.5],torchify=True)
        >>> k_2l = KernelGaussian(d=2,lengthscales=[.5,.5],torchify=True)
        >>> k_2l.double_integral_01d()
        tensor(0.5836, grad_fn=<MulBackward0>)
        >>> k_1l.double_integral_01d()
        tensor(0.5836, grad_fn=<MulBackward0>)
        >>> k_1l.single_integral_01d(x)
        tensor([0.5812, 0.4656, 0.5760, 0.5349], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        >>> k_2l.single_integral_01d(x)
        tensor([0.5812, 0.4656, 0.5760, 0.5349], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Derivatives 

        >>> kernel = KernelGaussian(
        ...     d = 3,
        ...     torchify = True,
        ...     scale = torch.from_numpy(rng.uniform(low=0,high=1,size=(1,))),
        ...     lengthscales = torch.from_numpy(rng.uniform(low=0,high=1,size=(3,))))
        >>> x0 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> x1 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> x2 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> x = torch.stack([x0,x1,x2],axis=-1)
        >>> z0 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> z1 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> z2 = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,))).requires_grad_(True)
        >>> z = torch.stack([z0,z1,z2],axis=-1)
        >>> c = torch.from_numpy(rng.uniform(low=0,high=1,size=(2,)))
        >>> beta0 = torch.tensor([
        ...     [1,0,0],
        ...     [0,2,0]])
        >>> beta1 = torch.tensor([
        ...     [0,0,2],
        ...     [2,1,0]])
        >>> with torch.no_grad():
        ...     y = kernel(x,z,beta0,beta1,c)
        >>> y
        tensor([ 97.6657,   3.8621, -65.9329,  -1.1932], dtype=torch.float64)
        >>> y_no_deriv = kernel(x,z)
        >>> y_first = y_no_deriv.clone()
        >>> y_first = torch.autograd.grad(y_first,x0,grad_outputs=torch.ones_like(y_first,requires_grad=True),create_graph=True)[0]
        >>> y_first = torch.autograd.grad(y_first,z2,grad_outputs=torch.ones_like(y_first,requires_grad=True),create_graph=True)[0]
        >>> y_first = torch.autograd.grad(y_first,z2,grad_outputs=torch.ones_like(y_first,requires_grad=True),create_graph=True)[0]
        >>> y_second = y_no_deriv.clone()
        >>> y_second = torch.autograd.grad(y_second,x1,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,x1,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,z0,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,z0,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> y_second = torch.autograd.grad(y_second,z1,grad_outputs=torch.ones_like(y_second,requires_grad=True),create_graph=True)[0]
        >>> yhat = (y_first*c[0]+y_second*c[1]).detach()
        >>> yhat
        tensor([ 97.6657,   3.8621, -65.9329,  -1.1932], dtype=torch.float64)
        >>> torch.allclose(y,yhat)
        True
    """

    def parsed___call__(self, x0, x1, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        k = scale*self.npt.exp(-((x0-x1)/(np.sqrt(2)*lengthscales))**2).prod(-1)
        return k
    
class KernelSquaredExponential(AbstractKernelGaussianSE):
    
    r"""
    Gaussian / Squared Exponential kernel implemented using the pairwise distance function.  
    Please use `KernelGaussian` when using derivative information.

    $$K(\boldsymbol{x},\boldsymbol{z}) = S \exp\left(-d_{\boldsymbol{\gamma}}^2(\boldsymbol{x},\boldsymbol{z})\right), \qquad d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z}) = \left\lVert\frac{\boldsymbol{x}-\boldsymbol{z}}{\sqrt{2}\boldsymbol{\gamma}}\right\rVert_2.$$

    Examples:
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelSquaredExponential(d=2)
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        array([1., 1., 1., 1.])
        >>> kernel(x[:,None,:],x[None,:,:])
        array([[1.        , 0.78888466, 0.9483142 , 0.8228498 ],
               [0.78888466, 1.        , 0.72380317, 0.62226176],
               [0.9483142 , 0.72380317, 1.        , 0.95613874],
               [0.8228498 , 0.62226176, 0.95613874, 1.        ]])
        
        Multiple randomizations 
        
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (6, 6, 5, 5)

        Batch hyperparameters 

        >>> kernel = KernelSquaredExponential(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1])
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        PyTorch

        >>> import torch
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelSquaredExponential(d=2,torchify=True)
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        tensor([1., 1., 1., 1.], dtype=torch.float64, grad_fn=<MulBackward0>)
        >>> kernel(x[:,None,:],x[None,:,:])
        tensor([[1.0000, 0.7889, 0.9483, 0.8228],
                [0.7889, 1.0000, 0.7238, 0.6223],
                [0.9483, 0.7238, 1.0000, 0.9561],
                [0.8228, 0.6223, 0.9561, 1.0000]], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Multiple randomizations 
        
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(6,5,2)))
        >>> kernel(x,x).shape 
        torch.Size([6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([6, 6, 5, 5])

        Batch hyperparameters 

        >>> kernel = KernelSquaredExponential(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1],
        ...     torchify = True)
        >>> kernel(x,x).shape 
        torch.Size([4, 3, 6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([4, 3, 6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([4, 3, 6, 6, 5, 5])
    """

    def parsed___call__(self, x0, x1, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        rdists = self.rel_pairwise_dist_func(x0,x1,lengthscales)
        k = scale*self.npt.exp(-rdists**2)
        return k

class KernelRationalQuadratic(AbstractKernelScaleLengthscales):

    r"""
    Rational Quadratic kernel 

    $$K(\boldsymbol{x},\boldsymbol{z}) = S \left(1+\frac{d_{\boldsymbol{\gamma}}^2(\boldsymbol{x},\boldsymbol{z})}{\alpha}\right)^{-\alpha}, \qquad d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z}) = \left\lVert\frac{\boldsymbol{x}-\boldsymbol{z}}{\sqrt{2}\boldsymbol{\gamma}}\right\rVert_2.$$

    Examples:
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelRationalQuadratic(d=2)
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        array([1., 1., 1., 1.])
        >>> kernel(x[:,None,:],x[None,:,:])
        array([[1.        , 0.80831912, 0.94960504, 0.83683297],
               [0.80831912, 1.        , 0.75572321, 0.67824456],
               [0.94960504, 0.75572321, 1.        , 0.95707312],
               [0.83683297, 0.67824456, 0.95707312, 1.        ]])
        
        Multiple randomizations 
        
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (6, 6, 5, 5)

        Batch hyperparameters 

        >>> kernel = KernelRationalQuadratic(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1])
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        PyTorch

        >>> import torch
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelRationalQuadratic(d=2,torchify=True)
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        tensor([1., 1., 1., 1.], dtype=torch.float64, grad_fn=<MulBackward0>)
        >>> kernel(x[:,None,:],x[None,:,:])
        tensor([[1.0000, 0.8083, 0.9496, 0.8368],
                [0.8083, 1.0000, 0.7557, 0.6782],
                [0.9496, 0.7557, 1.0000, 0.9571],
                [0.8368, 0.6782, 0.9571, 1.0000]], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Multiple randomizations 
        
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(6,5,2)))
        >>> kernel(x,x).shape 
        torch.Size([6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([6, 6, 5, 5])

        Batch hyperparameters 

        >>> kernel = KernelRationalQuadratic(
        ...     d = 2, 
        ...     shape_alpha = [4,3,1],
        ...     shape_lengthscales = [3,1],
        ...     torchify = True)
        >>> kernel(x,x).shape 
        torch.Size([4, 3, 6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([4, 3, 6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([4, 3, 6, 6, 5, 5])
    """
    
    AUTOGRADKERNEL = True

    def __init__(self,
            d, 
            scale = 1., 
            lengthscales = 1.,
            alpha = 1.,
            shape_scale = [1],
            shape_lengthscales = None, 
            shape_alpha = [1],
            tfs_scale = (tf_exp_eps_inv,tf_exp_eps),
            tfs_lengthscales = (tf_exp_eps_inv,tf_exp_eps),
            tfs_alpha = (tf_exp_eps_inv,tf_exp_eps),
            torchify = False, 
            requires_grad_scale = True, 
            requires_grad_lengthscales = True, 
            requires_grad_alpha = True, 
            device = "cpu",
            compile_call = False,
            comiple_call_kwargs = {},
            ):
        r"""
        Args:
            d (int): Dimension. 
            scale (Union[np.ndarray,torch.Tensor]): Scaling factor $S$.
            lengthscales (Union[np.ndarray,torch.Tensor]): Lengthscales $\boldsymbol{\gamma}$.
            alpha (Union[np.ndarray,torch.Tensor]): Scale mixture parameter $\alpha$.
            shape_scale (list): Shape of `scale` when `np.isscalar(scale)`. 
            shape_lengthscales (list): Shape of `lengthscales` when `np.isscalar(lengthscales)`
            shape_alpha (list): Shape of `alpha` when `np.isscalar(alpha)`
            tfs_scale (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_lengthscales (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            tfs_alpha (Tuple[callable,callable]): The first argument transforms to the raw value to be optimized; the second applies the inverse transform.
            torchify (bool): If `True`, use the `torch` backend. Set to `True` if computing gradients with respect to inputs and/or hyperparameters.
            requires_grad_scale (bool): If `True` and `torchify`, set `requires_grad=True` for `scale`.
            requires_grad_lengthscales (bool): If `True` and `torchify`, set `requires_grad=True` for `lengthscales`.
            requires_grad_alpha (bool): If `True` and `torchify`, set `requires_grad=True` for `alpha`.
            device (torch.device): If `torchify`, put things onto this device.
            compile_call (bool): If `True`, `torch.compile` the `parsed___call__` method. 
            comiple_call_kwargs (dict): When `compile_call` is `True`, pass these keyword arguments to `torch.compile`.
        """
        super().__init__(
            d = d, 
            scale = scale, 
            lengthscales = lengthscales,
            shape_scale = shape_scale,
            shape_lengthscales = shape_lengthscales, 
            tfs_scale = tfs_scale,
            tfs_lengthscales = tfs_lengthscales,
            torchify = torchify, 
            requires_grad_scale = requires_grad_scale, 
            requires_grad_lengthscales = requires_grad_lengthscales, 
            device = device,
            compile_call = compile_call,
            comiple_call_kwargs = comiple_call_kwargs,
        )
        self.raw_alpha = self.parse_assign_param(
            pname = "alpha",
            param = alpha, 
            shape_param = shape_alpha,
            requires_grad_param = requires_grad_alpha,
            tfs_param = tfs_alpha,
            endsize_ops = [1],
            constraints = ["POSITIVE"])
        self.tfs_alpha = tfs_alpha
        self.batch_param_names.append("alpha")
    
    @property
    def alpha(self):
        return self.tfs_alpha[1](self.raw_alpha)
    
    def parsed___call__(self, x0, x1, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        alpha = batch_params["alpha"][...,0]
        rdists = self.rel_pairwise_dist_func(x0,x1,lengthscales)
        k = scale*(1+rdists**2/alpha)**(-alpha)
        return k

class KernelMatern12(AbstractKernelScaleLengthscales):
    
    r""" 
    Matern kernel with $\alpha=1/2$. 

    $$K(\boldsymbol{x},\boldsymbol{z}) = S \exp\left(-d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z})\right), \qquad d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z}) = \left\lVert\frac{\boldsymbol{x}-\boldsymbol{z}}{\sqrt{2}\boldsymbol{\gamma}}\right\rVert_2.$$

    Examples:
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelMatern12(d=2)
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        array([1., 1., 1., 1.])
        >>> kernel(x[:,None,:],x[None,:,:])
        array([[1.        , 0.61448839, 0.79424131, 0.64302787],
               [0.61448839, 1.        , 0.56635268, 0.50219691],
               [0.79424131, 0.56635268, 1.        , 0.80913986],
               [0.64302787, 0.50219691, 0.80913986, 1.        ]])
        
        Multiple randomizations 
        
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (6, 6, 5, 5)

        Batch hyperparameters 

        >>> kernel = KernelMatern12(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1])
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        PyTorch

        >>> import torch
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelMatern12(d=2,torchify=True)
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        tensor([1., 1., 1., 1.], dtype=torch.float64, grad_fn=<MulBackward0>)
        >>> kernel(x[:,None,:],x[None,:,:])
        tensor([[1.0000, 0.6145, 0.7942, 0.6430],
                [0.6145, 1.0000, 0.5664, 0.5022],
                [0.7942, 0.5664, 1.0000, 0.8091],
                [0.6430, 0.5022, 0.8091, 1.0000]], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Multiple randomizations 
        
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(6,5,2)))
        >>> kernel(x,x).shape 
        torch.Size([6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([6, 6, 5, 5])

        Batch hyperparameters 

        >>> kernel = KernelMatern12(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1],
        ...     torchify = True)
        >>> kernel(x,x).shape 
        torch.Size([4, 3, 6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([4, 3, 6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([4, 3, 6, 6, 5, 5])
    """
    
    AUTOGRADKERNEL = True
    
    def parsed___call__(self, x0, x1, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        rdists = self.rel_pairwise_dist_func(x0,x1,lengthscales)
        k = scale*self.npt.exp(-rdists)
        return k

class KernelMatern32(AbstractKernelScaleLengthscales):
    
    r""" 
    Matern kernel with $\alpha=3/2$. 

    $$K(\boldsymbol{x},\boldsymbol{z}) = S \left(1+\sqrt{3} d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z})\right)\exp\left(-\sqrt{3}d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z})\right), \qquad d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z}) = \left\lVert\frac{\boldsymbol{x}-\boldsymbol{z}}{\sqrt{2}\boldsymbol{\gamma}}\right\rVert_2.$$

    Examples:
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelMatern32(d=2)
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        array([1., 1., 1., 1.])
        >>> kernel(x[:,None,:],x[None,:,:])
        array([[1.        , 0.79309639, 0.93871358, 0.82137958],
               [0.79309639, 1.        , 0.74137353, 0.66516872],
               [0.93871358, 0.74137353, 1.        , 0.9471166 ],
               [0.82137958, 0.66516872, 0.9471166 , 1.        ]])
        
        Multiple randomizations 
        
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (6, 6, 5, 5)

        Batch hyperparameters 

        >>> kernel = KernelMatern32(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1])
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        PyTorch

        >>> import torch
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelMatern32(d=2,torchify=True)
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        tensor([1., 1., 1., 1.], dtype=torch.float64, grad_fn=<MulBackward0>)
        >>> kernel(x[:,None,:],x[None,:,:])
        tensor([[1.0000, 0.7931, 0.9387, 0.8214],
                [0.7931, 1.0000, 0.7414, 0.6652],
                [0.9387, 0.7414, 1.0000, 0.9471],
                [0.8214, 0.6652, 0.9471, 1.0000]], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Multiple randomizations 
        
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(6,5,2)))
        >>> kernel(x,x).shape 
        torch.Size([6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([6, 6, 5, 5])

        Batch hyperparameters 

        >>> kernel = KernelMatern32(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1],
        ...     torchify = True)
        >>> kernel(x,x).shape 
        torch.Size([4, 3, 6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([4, 3, 6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([4, 3, 6, 6, 5, 5])
    """
    
    AUTOGRADKERNEL = True
    
    def parsed___call__(self, x0, x1, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        rdists = self.rel_pairwise_dist_func(x0,x1,lengthscales)
        k = scale*(1+np.sqrt(3)*rdists)*self.npt.exp(-np.sqrt(3)*rdists)
        return k
    
class KernelMatern52(AbstractKernelScaleLengthscales):
    
    r""" 
    Matern kernel with $\alpha=5/2$. 

    $$K(\boldsymbol{x},\boldsymbol{z}) = S \left(1+\sqrt{5} d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z}) + \frac{5}{3} d_{\boldsymbol{\gamma}}^2(\boldsymbol{x},\boldsymbol{z})\right)\exp\left(-\sqrt{5}d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z})\right), \qquad d_{\boldsymbol{\gamma}}(\boldsymbol{x},\boldsymbol{z}) = \left\lVert\frac{\boldsymbol{x}-\boldsymbol{z}}{\sqrt{2}\boldsymbol{\gamma}}\right\rVert_2.$$
    
    Examples:
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelMatern52(d=2)
        >>> x = rng.uniform(low=0,high=1,size=(4,2))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        array([1., 1., 1., 1.])
        >>> kernel(x[:,None,:],x[None,:,:])
        array([[1.        , 0.83612941, 0.95801903, 0.861472  ],
               [0.83612941, 1.        , 0.78812397, 0.71396963],
               [0.95801903, 0.78812397, 1.        , 0.96425994],
               [0.861472  , 0.71396963, 0.96425994, 1.        ]])
        
        Multiple randomizations 
        
        >>> x = rng.uniform(low=0,high=1,size=(6,5,2))
        >>> kernel(x,x).shape 
        (6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (6, 6, 5, 5)

        Batch hyperparameters 

        >>> kernel = KernelMatern52(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1])
        >>> kernel(x,x).shape 
        (4, 3, 6, 5)
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        (4, 3, 6, 5, 5)
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        (4, 3, 6, 6, 5, 5)

        PyTorch

        >>> import torch
        >>> rng = np.random.Generator(np.random.PCG64(7))
        >>> kernel = KernelMatern52(d=2,torchify=True)
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(4,2)))
        >>> kernel(x[0],x[0]).item()
        1.0
        >>> kernel(x,x)
        tensor([1., 1., 1., 1.], dtype=torch.float64, grad_fn=<MulBackward0>)
        >>> kernel(x[:,None,:],x[None,:,:])
        tensor([[1.0000, 0.8361, 0.9580, 0.8615],
                [0.8361, 1.0000, 0.7881, 0.7140],
                [0.9580, 0.7881, 1.0000, 0.9643],
                [0.8615, 0.7140, 0.9643, 1.0000]], dtype=torch.float64,
               grad_fn=<MulBackward0>)
        
        Multiple randomizations 
        
        >>> x = torch.from_numpy(rng.uniform(low=0,high=1,size=(6,5,2)))
        >>> kernel(x,x).shape 
        torch.Size([6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([6, 6, 5, 5])

        Batch hyperparameters 

        >>> kernel = KernelMatern52(
        ...     d = 2, 
        ...     shape_scale = [4,3,1],
        ...     shape_lengthscales = [3,1],
        ...     torchify = True)
        >>> kernel(x,x).shape 
        torch.Size([4, 3, 6, 5])
        >>> kernel(x[:,:,None,:],x[:,None,:,:]).shape
        torch.Size([4, 3, 6, 5, 5])
        >>> kernel(x[:,None,:,None,:],x[None,:,None,:,:]).shape
        torch.Size([4, 3, 6, 6, 5, 5])
    """
    
    AUTOGRADKERNEL = True
    
    def parsed___call__(self, x0, x1, batch_params):
        scale = batch_params["scale"][...,0]
        lengthscales = batch_params["lengthscales"]
        rdists = self.rel_pairwise_dist_func(x0,x1,lengthscales)
        k = scale*((1+np.sqrt(5)*rdists+5/3*rdists**2)*self.npt.exp(-np.sqrt(5)*rdists))
        return k
        



from setuptools import Extension

ext_modules = [
        Extension(
            name = 'qmcpy.discrete_distribution._c_lib._c_lib',
            sources = [
                'qmcpy/discrete_distribution/_c_lib/util.c',
                'qmcpy/discrete_distribution/_c_lib/halton_qrng.c',
            ]
        )
    ]

def pdm_build_update_setup_kwargs(context, setup_kwargs):
    setup_kwargs.update(ext_modules=ext_modules)